//****************************************************************
// Read/write xlsx (Office Open XML file format (Spreadsheet))
// Author:  Ruslan V. Neborak
// e-mail:  avemey@tut.by
// URL:     http://avemey.com
// License: zlib
// Last update: 2012.07.04
//----------------------------------------------------------------
{
 Copyright (C) 2012 Ruslan Neborak

  This software is provided 'as-is', without any express or implied
 warranty. In no event will the authors be held liable for any damages
 arising from the use of this software.

 Permission is granted to anyone to use this software for any purpose,
 including commercial applications, and to alter it and redistribute it
 freely, subject to the following restrictions:

    1. The origin of this software must not be misrepresented; you must not
    claim that you wrote the original software. If you use this software
    in a product, an acknowledgment in the product documentation would be
    appreciated but is not required.

    2. Altered source versions must be plainly marked as such, and must not be
    misrepresented as being the original software.

    3. This notice may not be removed or altered from any source
    distribution.
}
//****************************************************************
unit zexslx;

interface

{$I zexml.inc}
{$I compver.inc}

{$IFDEF FPC}
  {$mode objfpc}{$H+}
{$ENDIF}

uses
  SysUtils, classes, zsspxml, zexmlss, zesavecommon, graphics, zeformula
  {$IFNDEF FPC}
  ,windows
  {$ELSE}
  ,LCLType,
  LResources
  {$ENDIF}
  {$IFDEF FPC},zipper{$ENDIF};

type
  TZXLSXFileItem = record
    name: string;     //  
    original: string; // 
    ftype: integer;   // 
  end;

  TZXLSXFileArray = array of TZXLSXFileItem;

  TZXLSXRelations = record
    id: string;       //rID
    ftype: integer;   // 
    target: string;   //  
    fileid: integer;  //  
    name: string;     // 
    state: byte;      //
    sheetid: integer; // 
  end;

  TZXLSXRelationsArray = array of TZXLSXRelations;

function ReadXSLXPath(var XMLSS: TZEXMLSS; DirName: string): integer;

{$IFDEF FPC}
function ReadXSLX(var XMLSS: TZEXMLSS; FileName: string): integer;
{$ENDIF}

function SaveXmlssToXLSXPath(var XMLSS: TZEXMLSS; PathName: string; const SheetsNumbers:array of integer;
                             const SheetsNames: array of string; TextConverter: TAnsiToCPConverter; CodePageName: string; BOM: ansistring = ''): integer;

// ,     
function ZEXSLXReadContentTypes(var Stream: TStream; var FileArray: TZXLSXFileArray; var FilesCount: integer): boolean;
function ZEXSLXReadSharedStrings(var Stream: TStream; var StrArray: TZESaveStrArray; var StrCount: integer): boolean;
function ZEXSLXReadStyles(var XMLSS: TZEXMLSS; var Stream: TStream): boolean;
function ZE_XSLXReadRelationships(var Stream: TStream; var Relations: TZXLSXRelationsArray; var RelationsCount: integer; var isWorkSheet: boolean; needReplaceDelimiter: boolean): boolean;
function ZEXSLXReadWorkBook(var XMLSS: TZEXMLSS; var Stream: TStream; var Relations: TZXLSXRelationsArray; var RelationsCount: integer): boolean;
function ZEXSLXReadSheet(var XMLSS: TZEXMLSS; var Stream: TStream; const SheetName: string; var StrArray: TZESaveStrArray; StrCount: integer): boolean;

//     
function ZEXLSXCreateStyles(var XMLSS: TZEXMLSS; Stream: TStream; TextConverter: TAnsiToCPConverter; CodePageName: string; BOM: ansistring): integer;
function ZEXLSXCreateWorkBook(var XMLSS: TZEXMLSS; Stream: TStream; const _pages: TZESaveIntArray;
                              const _names: TZESaveStrArray; PageCount: integer; TextConverter: TAnsiToCPConverter; CodePageName: string; BOM: ansistring): integer;
function ZEXLSXCreateSheet(var XMLSS: TZEXMLSS; Stream: TStream; SheetNum: integer; TextConverter: TAnsiToCPConverter; CodePageName: string; BOM: ansistring; var isHaveComments: boolean): integer;
function ZEXLSXCreateContentTypes(var XMLSS: TZEXMLSS; Stream: TStream; PageCount: integer; CommentCount: integer; const PagesComments: TZESaveIntArray;
                                  TextConverter: TAnsiToCPConverter; CodePageName: string; BOM: ansistring): integer;
                             

implementation

type
  //
  TZEXLSXFont = record
    name: string;
    bold: boolean;
    italic: boolean;
    underline: boolean;
    strike: boolean;
    charset: integer;
    color: TColor;
    fontsize: integer;
  end;

  TZEXLSXFontArray = array of TZEXLSXFont; // 

{$IFDEF FPC}
  //   
  TXSLXZipHelper = class
  private
    FXMLSS: TZEXMLSS;
    FRetCode: integer;
    FFileArray: TZXLSXFileArray;
    FFilesCount: integer;
    FFileType: integer;
    FRelationsArray: array of TZXLSXRelationsArray;
    FRelationsCounts: array of integer;
    FRelationsCount: integer;
    FSheetRelationNumber: integer;
    FFileNumber: integer;
    FStrArray: TZESaveStrArray;
    FStrCount: integer;
    FListName: string;
    function GetFileItem(num: integer): TZXLSXFileItem;
    function GetRelationsCounts(num: integer): integer;
    function GetRelationsArray(num: integer): TZXLSXRelationsArray;
  protected
  public
    constructor Create(); virtual;
    destructor Destroy(); overload;
    procedure DoCreateOutZipStream(Sender: TObject; var AStream: TStream; AItem: TFullZipFileEntry);
    procedure DoDoneOutZipStream(Sender: TObject; var AStream: TStream; AItem: TFullZipFileEntry);
    property XMLSS: TZEXMLSS read FXMLSS write FXMLSS;                  //
    property ListName: string read FListName write FListName;           // 
    property RetCode: integer read FRetCode;                            // 
    property FileArray[num: integer]: TZXLSXFileItem read GetFileItem;  //
    property FilesCount: integer read FFilesCount;                      //- 
    property FileType: integer read FFileType write FFileType;          // 
    property FileNumber: integer read FFileNumber write FFileNumber;    //
    property RelationsCounts[num: integer]: integer read GetRelationsCounts; //-  
    property RelationsCount: integer read FRelationsCount;                   //- 
    property RelationsArray[num: integer]: TZXLSXRelationsArray read GetRelationsArray; //
    property SheetRelationNumber: integer read FSheetRelationNumber;        
  end;

constructor TXSLXZipHelper.Create();
begin
  inherited;
  FXMLSS := nil;
  FRetCode := 0;
  FFilesCount := 0;
  SetLength(FFileArray, 0);
  FFileType := -1;
  FRelationsCount := 0;
  FSheetRelationNumber := -1;
  SetLength(FRelationsArray, 0);
  SetLength(FRelationsCounts, 0);
  FFileNumber := -1;
  SetLength(FStrArray, 0);
  FStrCount := 0;
  FListName := '';
end;

destructor TXSLXZipHelper.Destroy();
var
  i: integer;

begin
  SetLength(FFileArray, 0);
  FFileArray := nil;
  for i := 0 to FRelationsCount - 1 do
  begin
    Setlength(FRelationsArray[i], 0);
    FRelationsArray[i] := nil;
  end;
  SetLength(FRelationsCounts, 0);
  FRelationsCounts := nil;
  SetLength(FRelationsArray, 0);
  FRelationsArray := nil;
  SetLength(FStrArray, 0);
  FStrArray := nil;
  inherited;
end;

function TXSLXZipHelper.GetFileItem(num: integer): TZXLSXFileItem;
begin
  if ((num >= 0) and (num < FilesCount)) then
    result := FFileArray[num];
end;

function TXSLXZipHelper.GetRelationsCounts(num: integer): integer;
begin
  if ((num >= 0) and (num < FRelationsCount)) then
    result := FRelationsCounts[num];
end;

function TXSLXZipHelper.GetRelationsArray(num: integer): TZXLSXRelationsArray;
begin
  if ((num >= 0) and (num < FRelationsCount)) then
    result := FRelationsArray[num];
end;

procedure TXSLXZipHelper.DoCreateOutZipStream(Sender: TObject; var AStream: TStream; AItem: TFullZipFileEntry);
begin
  AStream := TMemorystream.Create;
end;

procedure TXSLXZipHelper.DoDoneOutZipStream(Sender: TObject; var AStream: TStream; AItem: TFullZipFileEntry);
var
  b: boolean;
  j, k: integer;

begin
  if (Assigned(AStream)) then
  begin
    try
      AStream.Position := 0;
      // 
      if (FileType = -2) then
      begin
        if (not ZEXSLXReadContentTypes(AStream,  FFileArray, FFilesCount)) then
          FRetCode := FRetCode or 3;
        for j := 0 to FFilesCount - 1 do
        if (Length(FFileArray[j].name) > 0) then
          if (FFileArray[j].name[1] = '/') then
            delete(FFileArray[j].name, 1, 1);
      end else
      //relationships
      if (FileType = 3) then
      begin
        SetLength(FRelationsArray, FRelationsCount + 1);
        SetLength(FRelationsCounts, FRelationsCount + 1);

        if (not ZE_XSLXReadRelationships(AStream, FRelationsArray[FRelationsCount], FRelationsCounts[FRelationsCount], b, true)) then
        begin
          FRetCode := FRetCode or 4;
          exit;
        end;
        if (b) then
        begin
          FSheetRelationNumber := FRelationsCount;
          for j := 0 to FRelationsCounts[FRelationsCount] - 1 do
          if (FRelationsArray[FRelationsCount][j].ftype = 0) then
            for k := 0 to FilesCount - 1 do
            if (FRelationsArray[FRelationsCount][j].fileid < 0) then
            begin
              if ((pos(FRelationsArray[FRelationsCount][j].target, FFileArray[k].original)) > 0) then
              begin
                FRelationsArray[RelationsCount][j].fileid := k;
                break;
              end;
            end;
        end; //if
        inc(FRelationsCount);
      end else
      //sharedstrings
      if (FileType = 4) then
      begin
        if (not ZEXSLXReadSharedStrings(AStream, FStrArray, FStrCount)) then
          FRetCode := FRetCode or 3;
      end else
      //
      if (FileType = 1) then
      begin
        if (not ZEXSLXReadStyles(FXMLSS, AStream)) then
          FRetCode := FRetCode or 5;
      end else
      //Workbook
      if (FileType = 2) then
      begin
        if (not ZEXSLXReadWorkBook(FXMLSS, AStream, FRelationsArray[FSheetRelationNumber], FRelationsCounts[FSheetRelationNumber])) then
          FRetCode := FRetCode or 3;
      end else
      //Sheet
      if (FileType = 0) then
      begin
        if (not ZEXSLXReadSheet(FXMLSS, AStream, ListName, FStrArray, FStrCount)) then
          FRetCode := FRetCode or 4;
      end;
    finally
      FreeAndNil(AStream)
    end;
  end;
end;
{$ENDIF}

//BooleanToStr  XLSX //TODO:  
function XLSXBoolToStr(value: boolean): string;
begin
  if (value) then
    result := 'true'
  else
    result := 'false';
end;

//     [Content_Types].xml
//INPUT
//  var Stream: TStream             -  
//  var FileArray: TZXLSXFileArray  -  
//  var FilesCount: integer         - - 
//RETURN
//      boolean - true -   
function ZEXSLXReadContentTypes(var Stream: TStream; var FileArray: TZXLSXFileArray; var FilesCount: integer): boolean;
var
  xml: TZsspXMLReaderH;
  s: string;
  t: integer;

begin
  result := false;
  xml := nil;
  try
    xml := TZsspXMLReaderH.Create();
    if (xml.BeginReadStream(Stream) <> 0) then
      exit;
    FilesCount := 0;
    while (not xml.Eof()) do
    begin
      xml.ReadTag();
      if ((xml.TagName = 'Override') and (xml.TagType = 5)) then
      begin
        s := xml.Attributes.ItemsByName['PartName'];
        if (length(s) > 0) then
        begin
          SetLength(FileArray, FilesCount + 1);
          FileArray[FilesCount].name := s;
          FileArray[FilesCount].original := s;
          s := xml.Attributes.ItemsByName['ContentType'];
          t := -1;
          if (s = 'application/vnd.openxmlformats-officedocument.spreadsheetml.worksheet+xml') then
            t := 0
          else
          if (s = 'application/vnd.openxmlformats-officedocument.spreadsheetml.styles+xml') then
            t := 1
          else
          if (s = 'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet.main+xml') then
            t := 2
          else
          if (s = 'application/vnd.openxmlformats-package.relationships+xml') then
            t := 3
          else
          if (s = 'application/vnd.openxmlformats-officedocument.spreadsheetml.sharedStrings+xml') then
            t := 4
          else
          if (s = 'application/vnd.openxmlformats-package.core-properties+xml') then
            t := 5
          else
          if (s = 'application/vnd.openxmlformats-officedocument.spreadsheetml.comments+xml') then
            t := 6
          else
          if (s = 'application/vnd.openxmlformats-officedocument.vmlDrawing') then
            t := 7;

          FileArray[FilesCount].ftype := t;

          if (t > -1) then
            inc(FilesCount);
        end;
      end;
    end;

    result := true;
  finally
    if (Assigned(xml)) then
      FreeAndNil(xml);
  end;
end; //ZEXSLXReadContentTypes

//   sharedStrings.xml
//INPUT
//  var Stream: TStream           -   
//  var StrArray: TZESaveStrArray -    
//  var StrCount: integer         - - 
//RETURN
//      boolean - true -  
function ZEXSLXReadSharedStrings(var Stream: TStream; var StrArray: TZESaveStrArray; var StrCount: integer): boolean;
var
  xml: TZsspXMLReaderH;
  s: string;
  k: integer;

begin
  result := false;
  xml := nil;
  try
    xml := TZsspXMLReaderH.Create();
    if (xml.BeginReadStream(Stream) <> 0) then
      exit;
    StrCount := 0;

    while (not xml.Eof()) do
    begin
      xml.ReadTag();
      if ((xml.TagName = 'si') and (xml.TagType = 4)) then
      begin
        s := '';
        k := 0;
        while (not((xml.TagName = 'si') and (xml.TagType = 6))) and (not xml.Eof) do
        begin
          xml.ReadTag();
          if ((xml.TagName = 't') and (xml.TagType = 6)) then
          begin
            if (k > 1) then
              s := s + {$IFDEF FPC} LineEnding {$ELSE} sLineBreak {$ENDIF};
            s := s + xml.TextBeforeTag;
          end;
          if (xml.TagName = 'r') and (xml.TagType = 6) then
            inc(k);
        end; //while
        SetLength(StrArray, StrCount + 1);
        StrArray[StrCount] := s;
        inc(StrCount);
      end; //if
    end; //while

    result := true;

  finally
    if (Assigned(xml)) then
      FreeAndNil(xml);
  end;
end; //ZEXSLXReadSharedStrings

//  
//INPUT
//  var XMLSS: TZEXMLSS           - 
//  var Stream: TStream           -   
//  const SheetName: string       -  
//  var StrArray: TZESaveStrArray -   
//      StrCount: integer         - -  
//RETURN
//      boolean - true -   
function ZEXSLXReadSheet(var XMLSS: TZEXMLSS; var Stream: TStream; const SheetName: string; var StrArray: TZESaveStrArray; StrCount: integer): boolean;
var
  xml: TZsspXMLReaderH;
  _currPage: integer;
  _currRow: integer;
  _currCol: integer;
  _currSheet: TZSheet;
  _currCell: TZCell;
  s: string;
  _tmpr: real;
  _t: integer;

  // - 
  procedure CheckRow(RowCount: integer);
  begin
    if (_currSheet.RowCount < RowCount) then
      _currSheet.RowCount := RowCount;
  end;

  // - 
  procedure CheckCol(ColCount: integer);
  begin
    if (_currSheet.ColCount < ColCount) then
      _currSheet.ColCount := ColCount
  end;

  // /
  procedure _ReadSheetData();
  var
    t: integer;
    v: string;
    _num: integer;
    _type: string;
    _cr, _cc: integer;

  begin
    CheckRow(1);
    CheckCol(1);
    while (not ((xml.TagName = 'sheetData') and (xml.TagType = 6))) do
    begin
      xml.ReadTag();
      if (xml.Eof) then
        break;

      //
      if (xml.TagName = 'c') then
      begin
        s := xml.Attributes.ItemsByName['r']; //
        if (length(s) > 0) then
          if (ZEGetCellCoords(s, _cc, _cr)) then
          begin
            _currCol := _cc;
            CheckCol(_cc + 1);
          end;

        _type := xml.Attributes.ItemsByName['t']; //
        //s := xml.Attributes.ItemsByName['cm'];
        //s := xml.Attributes.ItemsByName['ph'];
        //s := xml.Attributes.ItemsByName['vm'];
        v := '';
        _num := 0;
        _currCell := _currSheet.Cell[_currCol, _currRow];
        s := xml.Attributes.ItemsByName['s']; //
        if (length(s) > 0) then
          if (tryStrToInt(s, t)) then
            _currCell.CellStyle := t;
        if (xml.TagType = 4) then
        while (not ((xml.TagName = 'c') and (xml.TagType = 6))) do
        begin
          xml.ReadTag();
          if (xml.Eof) then
            break;

          //is  

          if (((xml.TagName = 'v') or (xml.TagName = 't')) and (xml.TagType = 6)) then
          begin
            if (_num > 0) then
              v := v + {$IFDEF FPC} LineEnding {$ELSE} sLineBreak {$ENDIF};
            v := v + xml.TextBeforeTag;  
            inc(_num);
          end else
          if ((xml.TagName = 'f') and (xml.TagType = 6)) then
            _currCell.Formula := ZEReplaceEntity(xml.TextBeforeTag);

        end; //while

        //s - sharedstring
        //b - boolean
        //n - number
        //e - error
        //str - string
        //inlineStr - inline string ??
        if (_type = 's') then
        begin
          _currCell.CellType := ZEansistring;
          if (TryStrToInt(v, t)) then
            if ((t >= 0) and (t < StrCount)) then
              v := StrArray[t];
        end;

        _currCell.Data := v;
        inc(_currCol);
        CheckCol(_currCol + 1);
      end else
      //
      if ((xml.TagName = 'row') and (xml.TagType in [4, 5])) then
      begin
        _currCol := 0;
        s := xml.Attributes.ItemsByName['r']; // 
        if (length(s) > 0) then
          if (TryStrToInt(s, t)) then
          begin
            _currRow := t - 1;
            CheckRow(t);
          end;
        //s := xml.Attributes.ItemsByName['collapsed'];
        //s := xml.Attributes.ItemsByName['customFormat'];
        //s := xml.Attributes.ItemsByName['customHeight'];
        s := xml.Attributes.ItemsByName['hidden'];
        if (length(s) > 0) then
          _currSheet.Rows[_currRow].Hidden := ZEStrToBoolean(s);

        s := xml.Attributes.ItemsByName['ht']; // ?
        if (length(s) > 0) then
        begin
          _tmpr := ZETryStrToFloat(s, 10);
          _tmpr := _tmpr / 2.835; //???
          _currSheet.Rows[_currRow].HeightMM := _tmpr;
        end;

        //s := xml.Attributes.ItemsByName['outlineLevel'];
        //s := xml.Attributes.ItemsByName['ph'];

        s := xml.Attributes.ItemsByName['s']; // 
        if (length(s) > 0) then
          if (TryStrToInt(s, t)) then
          begin
            //   
          end;
        //s := xml.Attributes.ItemsByName['spans'];
        //s := xml.Attributes.ItemsByName['thickBot'];
        //s := xml.Attributes.ItemsByName['thickTop'];

        if (xml.TagType = 5) then
        begin
          inc(_currRow);
          CheckRow(_currRow + 1);
        end;
      end else
      // 
      if ((xml.TagName = 'row') and (xml.TagType = 6)) then
      begin
        inc(_currRow);
        CheckRow(_currRow + 1);
      end;
    end; //while
  end; //_ReadSheetData

  //  
  procedure _ReadMerge();
  var
    i, t, num: integer;
    x1, x2, y1, y2: integer;
    s1, s2: string;
    b: boolean;

    function _GetCoords(var x, y: integer): boolean;
    begin
      result := true;
      x := ZEGetColByA1(s1);
      if (x < 0) then
        result := false;
      if (not TryStrToInt(s2, y)) then
        result := false
      else
        dec(y);  
      b := result;
    end; //_GetCoords

  begin
    while (not ((xml.TagName = 'mergeCells') and (xml.TagType = 6))) do
    begin
      xml.ReadTag();
      if (xml.Eof) then
        break;

      if ((xml.TagName = 'mergeCell') and (xml.TagType in [4, 5])) then
      begin
        s := xml.Attributes.ItemsByName['ref'];
        t := length(s);
        if (t > 0) then
        begin
          s := s + ':';
          s1 := '';
          s2 := '';
          b := true;
          num := 0;
          for i := 1 to t + 1 do
          case s[i] of
            'A'..'Z', 'a'..'z': s1 := s1 + s[i];
            '0'..'9': s2 := s2 + s[i];
            ':':
              begin
                inc(num);
                if (num > 2) then
                begin
                  b := false;
                  break;
                end;
                if (num = 1) then
                begin
                  if (not _GetCoords(x1, y1)) then
                    break;
                end else
                begin
                  if (not _GetCoords(x2, y2)) then
                    break;
                end;
                s1 := '';
                s2 := '';
              end;
            else
            begin
              b := false;
              break;
            end;
          end; //case

          if (b) then
          begin
            CheckRow(y1 + 1);
            CheckRow(y2 + 1);
            CheckCol(x1 + 1);
            CheckCol(x2 + 1);
            _currSheet.MergeCells.AddRectXY(x1, y1, x2, y2);
          end;

        end; //if
      end; //if
    end; //while
  end; //_ReadMerge

  //
  procedure _ReadCols();
  var
    num: integer;
    t: real;

  begin
    num := 0;
    while (not ((xml.TagName = 'cols') and (xml.TagType = 6))) do
    begin
      xml.ReadTag();
      if (xml.Eof) then
        break;

      if ((xml.TagName = 'col') and (xml.TagType in [4, 5])) then
      begin
        CheckCol(num + 1);
        s := xml.Attributes.ItemsByName['bestFit'];
        if (length(s) > 0) then
          _currSheet.Columns[num].AutoFitWidth := ZETryStrToBoolean(s);

        //s := xml.Attributes.ItemsByName['collapsed'];
        //s := xml.Attributes.ItemsByName['customWidth'];
        s := xml.Attributes.ItemsByName['hidden'];
        if (length(s) > 0) then
          _currSheet.Columns[num].Hidden := ZETryStrToBoolean(s);

        //s := xml.Attributes.ItemsByName['max'];
        //s := xml.Attributes.ItemsByName['min'];
        //s := xml.Attributes.ItemsByName['outlineLevel'];
        //s := xml.Attributes.ItemsByName['phonetic'];
        s := xml.Attributes.ItemsByName['style'];

        s := xml.Attributes.ItemsByName['width'];
        if (length(s) > 0) then
        begin
          t := ZETryStrToFloat(s, 5.14509803921569);
          t := 10 * t / 5.14509803921569;
          _currSheet.Columns[num].WidthMM := t;
        end;

        inc(num);
      end; //if
    end; //while    
  end; //_ReadCols

  function _StrToMM(const st: string; var retFloat: real): boolean;
  begin
    result := false;
    if (length(s) > 0) then
    begin
      retFloat := ZETryStrToFloat(st, -1);
      if (retFloat > 0) then
      begin
        result := true;
        retFloat := retFloat * ZE_MMinInch;
      end;
    end;
  end; //_StrToMM

begin
  xml := nil;
  result := false;
  try
    xml := TZsspXMLReaderH.Create();
    if (xml.BeginReadStream(Stream) <> 0) then
      exit;

    _currPage := XMLSS.Sheets.Count;
    XMLSS.Sheets.Count := XMLSS.Sheets.Count + 1;
    _currRow := 0;
    _currSheet := XMLSS.Sheets[_currPage];
    _currSheet.Title := SheetName;

    while (not xml.Eof()) do
    begin
      xml.ReadTag();
      if ((xml.TagName = 'sheetData') and (xml.TagType = 4)) then
        _ReadSheetData()
      else
      if ((xml.TagName = 'mergeCells') and (xml.TagType = 4)) then
        _ReadMerge()
      else
      if ((xml.TagName = 'cols') and (xml.TagType = 4)) then
        _ReadCols()
      else
      //
      if ((xml.TagName = 'pageMargins') and (xml.TagType = 5)) then
      begin
        // 
        s := xml.Attributes.ItemsByName['bottom'];
        if (_StrToMM(s, _tmpr)) then
          _currSheet.SheetOptions.MarginBottom := round(_tmpr);
        s := xml.Attributes.ItemsByName['footer'];
        if (_StrToMM(s, _tmpr)) then
          _currSheet.SheetOptions.FooterMargin := round(_tmpr);
        s := xml.Attributes.ItemsByName['header'];
        if (_StrToMM(s, _tmpr)) then
          _currSheet.SheetOptions.HeaderMargin := round(_tmpr);
        s := xml.Attributes.ItemsByName['left'];
        if (_StrToMM(s, _tmpr)) then
          _currSheet.SheetOptions.MarginLeft := round(_tmpr);
        s := xml.Attributes.ItemsByName['right'];
        if (_StrToMM(s, _tmpr)) then
          _currSheet.SheetOptions.MarginRight := round(_tmpr);
        s := xml.Attributes.ItemsByName['top'];
        if (_StrToMM(s, _tmpr)) then
          _currSheet.SheetOptions.MarginTop := round(_tmpr);
      end else
      // 
      if ((xml.TagName = 'pageSetup') and (xml.TagType = 5)) then
      begin
        //s := xml.Attributes.ItemsByName['blackAndWhite'];
        //s := xml.Attributes.ItemsByName['cellComments'];
        //s := xml.Attributes.ItemsByName['copies'];
        //s := xml.Attributes.ItemsByName['draft'];
        //s := xml.Attributes.ItemsByName['errors'];
        s := xml.Attributes.ItemsByName['firstPageNumber'];
        if (length(s) > 0) then
          if (TryStrToInt(s, _t)) then
            _currSheet.SheetOptions.StartPageNumber := _t;
            
        //s := xml.Attributes.ItemsByName['fitToHeight'];
        //s := xml.Attributes.ItemsByName['fitToWidth'];
        //s := xml.Attributes.ItemsByName['horizontalDpi'];
        //s := xml.Attributes.ItemsByName['id'];
        s := xml.Attributes.ItemsByName['orientation'];
        if (length(s) > 0) then
        begin
          _currSheet.SheetOptions.PortraitOrientation := false;
          if (s = 'portrait') then
            _currSheet.SheetOptions.PortraitOrientation := true;
        end;
        
        //s := xml.Attributes.ItemsByName['pageOrder'];

        s := xml.Attributes.ItemsByName['paperSize'];
        if (length(s) > 0) then
          if (TryStrToInt(s, _t)) then
            _currSheet.SheetOptions.PaperSize := _t;
        //s := xml.Attributes.ItemsByName['paperHeight']; //  paperHeight  Width,  paperSize 
        //s := xml.Attributes.ItemsByName['paperWidth'];

        //s := xml.Attributes.ItemsByName['scale'];
        //s := xml.Attributes.ItemsByName['useFirstPageNumber'];
        //s := xml.Attributes.ItemsByName['usePrinterDefaults'];
        //s := xml.Attributes.ItemsByName['verticalDpi'];
      end else
      // 
      if ((xml.TagName = 'printOptions') and (xml.TagType = 5)) then
      begin
        //s := xml.Attributes.ItemsByName['gridLines'];
        //s := xml.Attributes.ItemsByName['gridLinesSet'];
        //s := xml.Attributes.ItemsByName['headings'];
        s := xml.Attributes.ItemsByName['horizontalCentered'];
        if (length(s) > 0) then
          _currSheet.SheetOptions.CenterHorizontal := ZEStrToBoolean(s);

        s := xml.Attributes.ItemsByName['verticalCentered'];
        if (length(s) > 0) then
          _currSheet.SheetOptions.CenterVertical := ZEStrToBoolean(s);
      end;
    end; //while
    
    result := true;
  finally
    if (Assigned(xml)) then
      FreeAndNil(xml);
  end;
end; //ZEXSLXReadSheet

//    (styles.xml)
//INPUT
//  var XMLSS: TZEXMLSS - 
//  var Stream: TStream - 
//RETURN
//      boolean - true -    
function ZEXSLXReadStyles(var XMLSS: TZEXMLSS; var Stream: TStream): boolean;
type

  TZXLSXBorderItem = record
    color: TColor;
    isColor: boolean;
    isEnabled: boolean;
    style: TZBorderType;
    width: integer;
  end;

  //   0 - left            
  //   1 - Top             
  //   2 - Right           
  //   3 - Bottom          
  //   4 - DiagonalLeft          
  //   5 - DiagonalRight         
  TZXLSXBorder = array[0..5] of TZXLSXBorderItem;
  TZXLSXBordersArray = array of TZXLSXBorder;

  TZXLSXCellAlignment = record
    horizontal: TZHorizontalAlignment;
    indent: integer;
    shrinkToFit: boolean;
    textRotation: integer;
    vertical: TZVerticalAlignment;
    wrapText: boolean;
  end;

  TZXLSXCellStyle = record
    applyAlignment: boolean;
    applyBorder: boolean;
    applyFont: boolean;
    applyProtection: boolean;
    borderId: integer;
    fillId: integer;
    fontId: integer;
    numFmtId: integer;
    xfId: integer;
    hidden: boolean;
    locked: boolean;
    alignment: TZXLSXCellAlignment;
  end;

  TZXLSXCellStylesArray = array of TZXLSXCellStyle;

  type TZXLSXStyle = record
    builtinId: integer;     //??
    customBuiltin: boolean; //??
    name: string;           //??
    xfId: integer;
  end;

  TZXLSXStyleArray = array of TZXLSXStyle;

  TZXLSXFill = record
    patternfill: TZCellPattern;
    bgcolor: TColor;
    patterncolor: TColor;
  end;

  TZXLSXFillArray = array of TZXLSXFill;

var
  xml: TZsspXMLReaderH;
  s: string;
  FontArray: TZEXLSXFontArray;
  FontCount: integer;
  BorderArray: TZXLSXBordersArray;
  BorderCount: integer;
  CellXfsArray: TZXLSXCellStylesArray;
  CellXfsCount: integer;
  CellStyleArray: TZXLSXCellStylesArray;
  CellStyleCount: integer;
  StyleArray: TZXLSXStyleArray;
  StyleCount: integer;
  FillArray: TZXLSXFillArray;
  FillCount: integer;
  _Style: TZStyle;
  t, i, n: integer;

  //   -
  //INPUT
  //  var fnt: TZEXLSXFont - 
  procedure ZEXLSXZeroFont(var fnt: TZEXLSXFont);
  begin
    fnt.name := 'Arial';
    fnt.bold := false;
    fnt.italic := false;
    fnt.underline := false;
    fnt.strike := false;
    fnt.charset := 204;
    fnt.color := clBlack;
    fnt.fontsize := 8;
  end; //ZEXLSXZeroFont

  // 
  //  var border: TZXLSXBorder - 
  procedure ZEXLSXZeroBorder(var border: TZXLSXBorder);
  var
    i: integer;

  begin
    for i := 0 to 5 do
    begin
      border[i].isColor := false;
      border[i].isEnabled := false;
      border[i].style := ZENone;
      border[i].width := 0;
    end;
  end; //ZEXLSXZeroBorder

  // 
  //INPUT
  //  var style: TZXLSXCellStyle -  XLSX
  procedure ZEXLSXZeroCellStyle(var style: TZXLSXCellStyle);
  begin
    style.applyAlignment := false;
    style.applyBorder := false;
    style.applyProtection := false;
    style.hidden := false;
    style.locked := false;
    style.borderId := -1;
    style.fontId := -1;
    style.fillId := -1;
    style.numFmtId := -1;
    style.xfId := -1;
    style.alignment.horizontal := ZHAutomatic;
    style.alignment.vertical := ZVAutomatic;
    style.alignment.shrinkToFit := false;
    style.alignment.wrapText := false;
    style.alignment.textRotation := 0;
    style.alignment.indent := 0;
  end; //ZEXLSXZeroCellStyle

  //TZEXLSXFont  TFont
  //INPUT
  //  var fnt: TZEXLSXFont  - XLSX 
  //  var font: TFont       -  
  procedure ZEXLSXFontToFont(var fnt: TZEXLSXFont; var font: TFont);
  begin
    if (Assigned(font)) then
    begin
      if (fnt.bold) then
        font.Style := font.Style + [fsBold];
      if (fnt.italic) then
        font.Style := font.Style + [fsItalic];
      if (fnt.underline) then
        font.Style := font.Style + [fsUnderline];
      if (fnt.strike) then
        font.Style := font.Style + [fsStrikeOut];
      font.Charset := fnt.charset;
      font.Name := fnt.name;
      font.Size := fnt.fontsize;
    end;
  end; //ZEXLSXFontToFont

  procedure _ReadFonts();
  var
    _currFont: integer;

  begin
    _currFont := -1;
    while (not ((xml.TagName = 'fonts') and (xml.TagType = 6))) do
    begin
      xml.ReadTag();
      if (xml.Eof) then
        break;

      s := xml.Attributes.ItemsByName['val'];
      if ((xml.TagName = 'font') and (xml.TagType = 4)) then
      begin
        _currFont := FontCount;
        inc(FontCount);
        SetLength(FontArray, FontCount);
        ZEXLSXZeroFont(FontArray[_currFont]);
      end else
      if (_currFont >= 0) then
      begin
        if ((xml.TagName = 'name') and (xml.TagType = 5)) then
          FontArray[_currFont].name := s
        else
        if ((xml.TagName = 'b') and (xml.TagType = 5)) then
          FontArray[_currFont].bold := ZEStrToBoolean(s)
        else
        if ((xml.TagName = 'charset') and (xml.TagType = 5)) then
        begin
          if (TryStrToInt(s, t)) then
            FontArray[_currFont].charset := t;
        end else
        if ((xml.TagName = 'color') and (xml.TagType = 5)) then
        begin
          s := xml.Attributes.ItemsByName['rgb'];
          if (length(s) > 2) then
            delete(s, 1, 2);
          if (length(s) > 0) then
            FontArray[_currFont].color := HTMLHexToColor(s);
        end else
        if ((xml.TagName = 'i') and (xml.TagType = 5)) then
          FontArray[_currFont].italic := ZEStrToBoolean(s)
        else
  //      if ((xml.TagName = 'outline') and (xml.TagType = 5)) then
  //      begin
  //      end else
        if ((xml.TagName = 'strike') and (xml.TagType = 5)) then
          FontArray[_currFont].strike := ZEStrToBoolean(s)
        else
        if ((xml.TagName = 'sz') and (xml.TagType = 5)) then
        begin
          if (TryStrToInt(s, t)) then
            FontArray[_currFont].fontsize := t;
        end else
        if ((xml.TagName = 'u') and (xml.TagType = 5)) then
        begin
          if (length(s) > 0) then
            if (s <> 'none') then
              FontArray[_currFont].underline := true;
        end
        else
        if ((xml.TagName = 'vertAlign') and (xml.TagType = 5)) then
        begin
        end;
      end; //if  
      //  
      //*b - bold
      //*charset
      //*color
      //?condense
      //?extend
      //?family
      //*i - italic
      //*name
      //?outline
      //?scheme
      //?shadow
      //*strike
      //*sz - size
      //*u - underline
      //*vertAlign

    end; //while
  end; //_ReadFonts

  procedure _ReadBorders();
  var
    _diagDown, _diagUP: boolean;
    _currBorder: integer; //  
    _currBorderItem: integer; //  (/ ...)
    _color: TColor;
    _isColor: boolean;

    procedure _SetCurBorder(borderNum: integer);
    var
      b: boolean;

    begin
      _currBorderItem := borderNum;
      s := xml.Attributes.ItemsByName['style'];
      if (Length(s) > 0) then
      begin
        b := true;
        BorderArray[_currBorder][borderNum].width := 1;
        if (s = 'thin') then
        begin
          BorderArray[_currBorder][borderNum].style := ZEContinuous;
        end else
        if (s = 'hair') then
          BorderArray[_currBorder][borderNum].style := ZEContinuous
        else
        if (s = 'dashed') then
          BorderArray[_currBorder][borderNum].style := ZEDash
        else
        if (s = 'dotted') then
          BorderArray[_currBorder][borderNum].style := ZEDot
        else
        if (s = 'dashDot') then
          BorderArray[_currBorder][borderNum].style := ZEDashDot
        else
        if (s = 'dashDotDot') then
          BorderArray[_currBorder][borderNum].style := ZEDashDotDot
        else
        if (s = 'slantDashDot') then
          BorderArray[_currBorder][borderNum].style := ZESlantDashDot
        else
        if (s = 'double') then
          BorderArray[_currBorder][borderNum].style := ZEDouble
        else
        if (s = 'medium') then
        begin
          BorderArray[_currBorder][borderNum].style := ZEContinuous;
          BorderArray[_currBorder][borderNum].width := 2;
        end else
        if (s = 'thick') then
        begin
          BorderArray[_currBorder][borderNum].style := ZEContinuous;
          BorderArray[_currBorder][borderNum].width := 3;
        end else
        if (s = 'mediumDashed') then
        begin
          BorderArray[_currBorder][borderNum].style := ZEDash;
          BorderArray[_currBorder][borderNum].width := 2;
        end else
        if (s = 'mediumDashDot') then
        begin
          BorderArray[_currBorder][borderNum].style := ZEDashDot;
          BorderArray[_currBorder][borderNum].width := 2;
        end else
        if (s = 'mediumDashDotDot') then
        begin
          BorderArray[_currBorder][borderNum].style := ZEDashDotDot;
          BorderArray[_currBorder][borderNum].width := 2;
        end else
        if (s = 'none') then
          BorderArray[_currBorder][borderNum].style := ZENone
        else
          b := false;

        BorderArray[_currBorder][borderNum].isEnabled := b;
      end;
    end; //_SetCurBorder

  begin
    _currBorderItem := -1;
    _diagDown := false;
    _diagUP := false;
    _color := clBlack;
    while (not ((xml.TagName = 'borders') and (xml.TagType = 6))) do
    begin
      xml.ReadTag();
      if (xml.Eof()) then
        break;

      if ((xml.TagName = 'border') and (xml.TagType = 4)) then
      begin
        _currBorder := BorderCount;
        inc(BorderCount);
        SetLength(BorderArray, BorderCount);
        ZEXLSXZeroBorder(BorderArray[_currBorder]);
        _diagDown := false;
        _diagUP := false;
        s := xml.Attributes.ItemsByName['diagonalDown'];
        if (length(s) > 0) then
          _diagDown := ZEStrToBoolean(s);

        s := xml.Attributes.ItemsByName['diagonalUp'];
        if (length(s) > 0) then
          _diagUP := ZEStrToBoolean(s);
      end else
      begin
        if (xml.TagType in [4, 5]) then
        begin
          if (xml.TagName = 'left') then
          begin
            _SetCurBorder(0);
          end else
          if (xml.TagName = 'right') then
          begin
            _SetCurBorder(2);
          end else
          if (xml.TagName = 'top') then
          begin
            _SetCurBorder(1);
          end else
          if (xml.TagName = 'bottom') then
          begin
            _SetCurBorder(3);
          end else
          if (xml.TagName = 'diagonal') then
          begin
            if (_diagUp) then
              _SetCurBorder(5);
            if (_diagDown) then
            begin
              if (_diagUp) then
                BorderArray[_currBorder][4] := BorderArray[_currBorder][5]
              else
                _SetCurBorder(4);
            end;
          end else
          //??  {
          if (xml.TagName = 'end') then
          begin
          end else
          if (xml.TagName = 'start') then
          begin
          end else
          //??  }
          if (xml.TagName = 'color') then
          begin
            _isColor := false;
            s := xml.Attributes.ItemsByName['rgb'];
            if (length(s) > 2) then
              delete(s, 1, 2);
            if (length(s) > 0) then
            begin
              _color := HTMLHexToColor(s);
              _isColor := true;
            end;
            if (_isColor and (_currBorderItem >= 0) and (_currBorderItem < 6)) then
            begin
              BorderArray[_currBorder][_currBorderItem].color := _color;
              BorderArray[_currBorder][_currBorderItem].isColor := true;
            end;
          end;
        end; //if
      end; //else

    end; //while
  end; //_ReadBorders

  procedure _ReadFills();
  var
    _currFill: integer;

  begin
    _currFill := -1;
    while (not ((xml.TagName = 'fills') and (xml.TagType = 6))) do
    begin
      xml.ReadTag();
      if (xml.Eof()) then
        break;

      if ((xml.TagName = 'fill') and (xml.TagType = 4)) then
      begin
        _currFill := FillCount;
        inc(FillCount);
        SetLength(FillArray, FillCount);
        FillArray[_currFill].patternfill := ZPNone;
        FillArray[_currFill].bgcolor := clWindow;
        FillArray[_currFill].patterncolor := clWindow;
      end else
      if ((xml.TagName = 'patternFill') and (xml.TagType in [4, 5])) then
      begin
        if (_currFill >= 0) then
        begin
          s := xml.Attributes.ItemsByName['patternType'];
          {
          *none	None
          *solid	Solid
          mediumGray	Medium Gray
          darkGray	Dary Gray
          lightGray	Light Gray
          darkHorizontal	Dark Horizontal
          darkVertical	Dark Vertical
          darkDown	Dark Down
          darkUpDark Up
          darkGrid	Dark Grid
          darkTrellis	Dark Trellis
          lightHorizontal	Light Horizontal
          lightVertical	Light Vertical
          lightDown	Light Down
          lightUp	Light Up
          lightGrid	Light Grid
          lightTrellis	Light Trellis
          *gray125	Gray 0.125
          *gray0625	Gray 0.0625
          }
          if (length(s) > 0) then
          begin
            if (s = 'solid') then
              FillArray[_currFill].patternfill := ZPSolid
            else
            if (s = 'none') then
              FillArray[_currFill].patternfill := ZPNone
            else
            if (s = 'gray125') then
              FillArray[_currFill].patternfill := ZPGray125
            else
            if (s = 'gray0625') then
              FillArray[_currFill].patternfill := ZPGray0625
            else
              FillArray[_currFill].patternfill := ZPSolid; //{tut}     
          end;
        end;
      end else
      if ((xml.TagName = 'fgColor') and (xml.TagType = 5)) then
      begin
        if (_currFill >= 0) then
        begin
          s := xml.Attributes.ItemsByName['rgb'];
          if (length(s) > 2) then
            delete(s, 1, 2);
          if (length(s) > 0) then
            FillArray[_currFill].patterncolor := HTMLHexToColor(s);
        end;
      end else
      if ((xml.TagName = 'bgColor') and (xml.TagType = 5)) then
      begin
        if (_currFill >= 0) then
        begin
          s := xml.Attributes.ItemsByName['rgb'];
          if (length(s) > 2) then
            delete(s, 1, 2);
          if (length(s) > 0) then
            FillArray[_currFill].bgcolor := HTMLHexToColor(s);
        end;
      end;
    end; //while
  end; //_ReadFills

  //  (cellXfs  cellStyleXfs)
  //INPUT
  //  const TagName: string           -  
  //  var CSA: TZXLSXCellStylesArray  -   
  //  var StyleCount: integer         - - 
  procedure _ReadCellCommonStyles(const TagName: string; var CSA: TZXLSXCellStylesArray; var StyleCount: integer);
  var
    _currCell: integer;
    b: boolean;

  begin
    _currCell := -1;
    while (not ((xml.TagName = TagName) and (xml.TagType = 6))) do
    begin
      xml.ReadTag();
      if (xml.Eof()) then
        break;

      b := false;  

      if ((xml.TagName = 'xf') and (xml.TagType = 4)) then
      begin
        _currCell := StyleCount;
        inc(StyleCount);
        SetLength(CSA, StyleCount);
        ZEXLSXZeroCellStyle(CSA[_currCell]);
        s := xml.Attributes.ItemsByName['applyAlignment'];
        if (length(s) > 0) then
          CSA[_currCell].applyAlignment := ZEStrToBoolean(s);

        s := xml.Attributes.ItemsByName['applyBorder'];
        if (length(s) > 0) then
          CSA[_currCell].applyBorder := ZEStrToBoolean(s)
        else
          b := true;

        s := xml.Attributes.ItemsByName['applyFont'];
        if (length(s) > 0) then
          CSA[_currCell].applyFont := ZEStrToBoolean(s);

        s := xml.Attributes.ItemsByName['applyProtection'];
        if (length(s) > 0) then
          CSA[_currCell].applyProtection := ZEStrToBoolean(s);

        s := xml.Attributes.ItemsByName['borderId'];
        if (length(s) > 0) then
          if (TryStrToInt(s, t)) then
          begin
            CSA[_currCell].borderId := t;
            if (b and (t >= 1)) then
              CSA[_currCell].applyBorder := true;
          end;

        s := xml.Attributes.ItemsByName['fillId'];
        if (length(s) > 0) then
          if (TryStrToInt(s, t)) then
            CSA[_currCell].fillId := t;

        s := xml.Attributes.ItemsByName['fontId'];
        if (length(s) > 0) then
          if (TryStrToInt(s, t)) then
            CSA[_currCell].fontId := t;

        s := xml.Attributes.ItemsByName['numFmtId'];
        if (length(s) > 0) then
          if (TryStrToInt(s, t)) then
            CSA[_currCell].numFmtId := t;

        s := xml.Attributes.ItemsByName['xfId'];
        if (length(s) > 0) then
          if (TryStrToInt(s, t)) then
            CSA[_currCell].xfId := t;
      end else
      if ((xml.TagName = 'alignment') and (xml.TagType = 5)) then
      begin
        if (_currCell >= 0) then
        begin
          s := xml.Attributes.ItemsByName['horizontal'];
          if (length(s) > 0) then
          begin
            if (s = 'left') then
              CSA[_currCell].alignment.horizontal := ZHLeft
            else
            if (s = 'right') then
              CSA[_currCell].alignment.horizontal := ZHRight
            else
            if ((s = 'center') or (s = 'centerContinuous')) then
              CSA[_currCell].alignment.horizontal := ZHCenter
            else
            if (s = 'fill') then
              CSA[_currCell].alignment.horizontal := ZHFill
            else
            if (s = 'justify') then
              CSA[_currCell].alignment.horizontal := ZHJustify
            else
            if (s = 'distributed') then
              CSA[_currCell].alignment.horizontal := ZHDistributed;
          end;

          s := xml.Attributes.ItemsByName['indent'];
          if (length(s) > 0) then
            if (TryStrToInt(s, t)) then
              CSA[_currCell].alignment.indent := t;

          s := xml.Attributes.ItemsByName['shrinkToFit'];
          if (length(s) > 0) then
            CSA[_currCell].alignment.shrinkToFit := ZEStrToBoolean(s);

          s := xml.Attributes.ItemsByName['textRotation'];
          if (length(s) > 0) then
            if (TryStrToInt(s, t)) then
              CSA[_currCell].alignment.textRotation := t;

          s := xml.Attributes.ItemsByName['vertical'];
          if (length(s) > 0) then
          begin
            if (s = 'center') then
              CSA[_currCell].alignment.vertical := ZVCenter
            else
            if (s = 'top') then
              CSA[_currCell].alignment.vertical := ZVTop
            else
            if (s = 'bottom') then
              CSA[_currCell].alignment.vertical := ZVBottom
            else
            if (s = 'justify') then
              CSA[_currCell].alignment.vertical := ZVJustify
            else
            if (s = 'distributed') then
              CSA[_currCell].alignment.vertical := ZVDistributed;
          end;

          s := xml.Attributes.ItemsByName['wrapText'];
          if (length(s) > 0) then
            CSA[_currCell].alignment.wrapText := ZEStrToBoolean(s);
        end; //if
      end else
      if ((xml.TagName = 'protection') and (xml.TagType = 5)) then
      begin
        if (_currCell >= 0) then
        begin
          s := xml.Attributes.ItemsByName['hidden'];
          if (length(s) > 0) then
            CSA[_currCell].hidden := ZEStrToBoolean(s);

          s := xml.Attributes.ItemsByName['locked'];
          if (length(s) > 0) then
            CSA[_currCell].locked := ZEStrToBoolean(s);
        end;
      end;
    end; //while
  end; //_ReadCellCommonStyles

  //  ?? (    ?)
  procedure _ReadCellStyles();
  var
    b: boolean;

  begin
    while (not ((xml.TagName = 'cellStyles') and (xml.TagType = 6))) do
    begin
      xml.ReadTag();
      if (xml.Eof()) then
        break;

      if ((xml.TagName = 'cellStyle') and (xml.TagType = 5)) then
      begin
        b := false;
        SetLength(StyleArray, StyleCount + 1);
        s := xml.Attributes.ItemsByName['builtinId']; //?
        if (length(s) > 0) then
          if (TryStrToInt(s, t)) then
            StyleArray[StyleCount].builtinId := t;

        s := xml.Attributes.ItemsByName['customBuiltin']; //?
        if (length(s) > 0) then
          StyleArray[StyleCount].customBuiltin := ZEStrToBoolean(s);

        s := xml.Attributes.ItemsByName['name']; //?
          StyleArray[StyleCount].name := s;

        s := xml.Attributes.ItemsByName['xfId'];
        if (length(s) > 0) then
          if (TryStrToInt(s, t)) then
          begin
            StyleArray[StyleCount].xfId := t;
            b := true;
          end;

        if (b) then
          inc(StyleCount);
      end;
    end; //while
  end; //_ReadCellStyles

  procedure _ReadColors();
  begin
    while (not ((xml.TagName = 'colors') and (xml.TagType = 6))) do
    begin
      xml.ReadTag();
      if (xml.Eof()) then
        break;

      if ((xml.TagName = 'rgbColor') and (xml.TagType = 5)) then
      begin
        s := xml.Attributes.ItemsByName['rgb'];
        if (length(s) > 2) then
          delete(s, 1, 2);
        if (length(s) > 0) then
        begin

        end;
      end;
    end; //while
  end; //_ReadColors

  // 
  //INPUT
  //  var XMLSSStyle: TZStyle         -   
  //  var XLSXStyle: TZXLSXCellStyle  -   xlsx
  procedure _ApplyStyle(var XMLSSStyle: TZStyle; var XLSXStyle: TZXLSXCellStyle);
  var
    i: integer;

  begin
    if (XLSXStyle.applyAlignment) then
    begin
      XMLSSStyle.Alignment.Horizontal := XLSXStyle.alignment.horizontal;
      XMLSSStyle.Alignment.Vertical := XLSXStyle.alignment.vertical;
      XMLSSStyle.Alignment.Indent := XLSXStyle.alignment.indent;
      XMLSSStyle.Alignment.Rotate := XLSXStyle.alignment.textRotation;
      XMLSSStyle.Alignment.ShrinkToFit := XLSXStyle.alignment.shrinkToFit;
      XMLSSStyle.Alignment.WrapText := XLSXStyle.alignment.wrapText;
    end;

    if (XLSXStyle.applyBorder) then
    begin
      n := XLSXStyle.borderId;
      if ((n >= 0) and (n < BorderCount)) then
        for i := 0 to 5 do
        if (BorderArray[n][i].isEnabled) then
        begin
          XMLSSStyle.Border[i].LineStyle := BorderArray[n][i].style;
          XMLSSStyle.Border[i].Weight := BorderArray[n][i].width;
          if (BorderArray[n][i].isColor) then
            XMLSSStyle.Border[i].Color := BorderArray[n][i].color;
        end;
    end;

    if (XLSXStyle.applyFont) then
    begin
      n := XLSXStyle.fontId;
      if ((n >= 0) and (n < FontCount)) then
      begin
        XMLSSStyle.Font.Name := FontArray[n].name;
        XMLSSStyle.Font.Size := FontArray[n].fontsize;
        XMLSSStyle.Font.Charset := FontArray[n].charset;
        XMLSSStyle.Font.Color := FontArray[n].color;
        if (FontArray[n].bold) then
          XMLSSStyle.Font.Style := [fsBold];
        if (FontArray[n].underline) then
          XMLSSStyle.Font.Style := XMLSSStyle.Font.Style + [fsUnderline];
        if (FontArray[n].italic) then
          XMLSSStyle.Font.Style := XMLSSStyle.Font.Style + [fsItalic];
        if (FontArray[n].strike) then
          XMLSSStyle.Font.Style := XMLSSStyle.Font.Style + [fsStrikeOut];
      end;
    end;

    if (XLSXStyle.applyProtection) then
    begin
      XMLSSStyle.Protect := XLSXStyle.locked;
      XMLSSStyle.HideFormula := XLSXStyle.hidden;
    end;

    n := XLSXStyle.fillId;
    if ((n >= 0) and (n < FillCount)) then
    begin
      XMLSSStyle.CellPattern := FillArray[n].patternfill;
      XMLSSStyle.BGColor := FillArray[n].bgcolor;
      XMLSSStyle.PatternColor := FillArray[n].patterncolor;
    end;
  end; //_ApplyStyle

begin
  result := false;
  xml := nil;
  try
    xml := TZsspXMLReaderH.Create();
    if (xml.BeginReadStream(Stream) <> 0) then
      exit;

    FontCount := 0;
    BorderCount := 0;
    CellStyleCount := 0;
    StyleCount := 0;
    CellXfsCount := 0;
    FillCount := 0;

    while (not xml.Eof()) do
    begin
      xml.ReadTag();

      if ((xml.TagName = 'fonts') and (xml.TagType = 4)) then
        _ReadFonts()
      else
      if ((xml.TagName = 'borders') and (xml.TagType = 4)) then
        _ReadBorders()
      else
      if ((xml.TagName = 'fills') and (xml.TagType = 4)) then
        _ReadFills()
      else
      //TODO: ,   cellStyleXfs  cellXfs.   cellStyles?
      if ((xml.TagName = 'cellStyleXfs') and (xml.TagType = 4)) then
        _ReadCellCommonStyles('cellStyleXfs', CellStyleArray, CellStyleCount)//_ReadCellStyleXfs()
      else
      if ((xml.TagName = 'cellXfs') and (xml.TagType = 4)) then  // ?
        _ReadCellCommonStyles('cellXfs', CellXfsArray, CellXfsCount) //_ReadCellXfs()
      else
      if ((xml.TagName = 'cellStyles') and (xml.TagType = 4)) then //??
        _ReadCellStyles()
      else
      if ((xml.TagName = 'colors') and (xml.TagType = 4)) then
        _ReadColors();
    end; //while

    //    ,   
    //{tut}

    XMLSS.Styles.Count := CellXfsCount;
    for i := 0 to CellXfsCount - 1 do
    begin
      t := CellXfsArray[i].xfId;
      _Style := XMLSS.Styles[i];
      if ((t >= 0) and (t < CellStyleCount)) then
        _ApplyStyle(_Style, CellStyleArray[t]);
      _ApplyStyle(_Style, CellXfsArray[i]);
    end;

    result := true;
  finally
    if (Assigned(xml)) then
      FreeAndNil(xml);
    SetLength(FontArray, 0);
    FontArray := nil;
    SetLength(BorderArray, 0);
    BorderArray := nil;
    SetLength(CellStyleArray, 0);
    CellStyleArray := nil;
    SetLength(StyleArray, 0);
    StyleArray := nil;
    SetLength(CellXfsArray, 0);
    CellXfsArray := nil;
    SetLength(FillArray, 0);
    FillArray := nil;
  end;
end; //ZEXSLXReadStyles

//   (workbook.xml)
//INPUT
//  var XMLSS: TZEXMLSS                 - 
//  var Stream: TStream                 - 
//  var Relations: TZXLSXRelationsArray - 
//  var RelationsCount: integer         - -
//RETURN
//      boolean - true -    
function ZEXSLXReadWorkBook(var XMLSS: TZEXMLSS; var Stream: TStream; var Relations: TZXLSXRelationsArray; var RelationsCount: integer): boolean;
var
  xml: TZsspXMLReaderH;
  s: string;
  i: integer;
  t: integer;

begin
  result := false;
  xml := nil;
  try
    xml := TZsspXMLReaderH.Create();
    if (xml.BeginReadStream(Stream) <> 0) then
      exit;

    while (not xml.Eof()) do
    begin
      xml.ReadTag();

      if ((xml.TagName = 'sheet') and (xml.TagType = 5)) then
      begin
        s := xml.Attributes.ItemsByName['r:id'];
        for i := 0 to RelationsCount - 1 do
          if (Relations[i].id = s) then
          begin
            Relations[i].name := xml.Attributes.ItemsByName['name'];
            s := xml.Attributes.ItemsByName['sheetId'];
            relations[i].sheetid := -1;
            if (TryStrToInt(s, t)) then
              relations[i].sheetid := t;
            s := xml.Attributes.ItemsByName['state'];
            break;
          end;
      end else
      if ((xml.TagName = 'workbookView') and (xml.TagType = 5)) then
      begin
        s := xml.Attributes.ItemsByName['activeTab'];
        s := xml.Attributes.ItemsByName['firstSheet'];
        s := xml.Attributes.ItemsByName['showHorizontalScroll'];
        s := xml.Attributes.ItemsByName['showSheetTabs'];
        s := xml.Attributes.ItemsByName['showVerticalScroll'];
        s := xml.Attributes.ItemsByName['tabRatio'];
        s := xml.Attributes.ItemsByName['windowHeight'];
        s := xml.Attributes.ItemsByName['windowWidth'];
        s := xml.Attributes.ItemsByName['xWindow'];
        s := xml.Attributes.ItemsByName['yWindow'];
      end;
    end; //while
    result := true;
  finally
    if (Assigned(xml)) then
      FreeAndNil(xml);
  end;
end; //ZEXSLXReadWorkBook

//   +     
//INPUT
//  var FileArray: TZXLSXFileArray  - 
//      FilesCount: integer         - - 
procedure ZE_XSLXReplaceDelimiter(var FileArray: TZXLSXFileArray; FilesCount: integer);
var
  i, j, k: integer;

begin
  for i := 0 to FilesCount - 1 do
  begin
    k := length(FileArray[i].name);
    if (k > 1) then
    begin
      if (FileArray[i].name[1] = '/') then
        Delete(FileArray[i].name, 1, 1);
      if (PathDelim <> '/') then
        for j := 1 to k - 1 do
          if (FileArray[i].name[j] = '/') then
            FileArray[i].name[j] := PathDelim;
    end;
  end;
end; //ZE_XSLXReplaceDelimiter

//  /  (*.rels: workbook.xml.rels  .rels)
//INPUT
//  var Stream: TStream                 -   
//  var Relations: TZXLSXRelationsArray -   
//  var RelationsCount: integer         - -
//  var isWorkSheet: boolean            -  workbook.xml.rels
//      needReplaceDelimiter: boolean   -    
//RETURN
//      boolean - true -  
function ZE_XSLXReadRelationships(var Stream: TStream; var Relations: TZXLSXRelationsArray; var RelationsCount: integer; var isWorkSheet: boolean; needReplaceDelimiter: boolean): boolean;
var
  xml: TZsspXMLReaderH;
  s: string;
  t: integer;

begin
  result := false;
  xml := nil;
  RelationsCount := 0;
  isWorkSheet := false;
  try
    xml := TZsspXMLReaderH.Create();
    if (xml.BeginReadStream(Stream) <> 0) then
      exit;

    while (not xml.Eof()) do
    begin
      xml.ReadTag();

      if ((xml.TagName = 'Relationship') and (xml.TagType = 5)) then
      begin
        SetLength(Relations, RelationsCount + 1);
        Relations[RelationsCount].id := xml.Attributes.ItemsByName['Id'];

        s := xml.Attributes.ItemsByName['Type'];
        t := -1;
        if (s = 'http://schemas.openxmlformats.org/officeDocument/2006/relationships/worksheet') then
          t := 0
        else
        if (s = 'http://schemas.openxmlformats.org/officeDocument/2006/relationships/styles') then
          t := 1
        else
        if (s = 'http://schemas.openxmlformats.org/officeDocument/2006/relationships/sharedStrings') then
          t := 2
        else
        if (s = 'http://schemas.openxmlformats.org/officeDocument/2006/relationships/officeDocument') then
          t := 3
        else
        if (s = 'http://schemas.openxmlformats.org/package/2006/relationships/metadata/core-properties') then
          t := 4
        else
        if (s = 'http://schemas.openxmlformats.org/officeDocument/2006/relationships/extended-properties') then
          t := 5;

        if ((t >= 0) and (t < 3)) then
          isWorkSheet := true;

        Relations[RelationsCount].fileid := -1;
        Relations[RelationsCount].state := 0;
        Relations[RelationsCount].sheetid := 0;
        Relations[RelationsCount].name := '';

        Relations[RelationsCount].ftype := t;
        Relations[RelationsCount].target := xml.Attributes.ItemsByName['Target'];
        if (t >= 0) then
          inc(RelationsCount);
      end;
    end; //while
    result := true;
  finally
    if (Assigned(xml)) then
      FreeAndNil(xml);
  end;
end; //ZE_XSLXReadRelationsips

//  xlsx
//INPUT
//  var XMLSS: TZEXMLSS - 
//  DirName: string     -  
//RETURN
//      integer -   (0 -  OK)
function ReadXSLXPath(var XMLSS: TZEXMLSS; DirName: string): integer;
var
  stream: TStream;
  FileArray: TZXLSXFileArray;
  FilesCount: integer;
  StrArray: TZESaveStrArray;
  StrCount: integer;
  RelationsArray: array of TZXLSXRelationsArray;
  RelationsCounts: array of integer;
  RelationsCount: integer;
  SheetRelationNumber: integer;
  i, j, k: integer;
  s: string;
  b: boolean;

begin
  result := 0;

  if (not ZE_CheckDirExist(DirName)) then
  begin
    result := -1;
    exit;
  end;

  XMLSS.Styles.Clear();
  XMLSS.Sheets.Count := 0;
  stream := nil;
  RelationsCount := 0;

  try
    try
      stream := TFileStream.Create(DirName + '[Content_Types].xml', fmOpenRead or fmShareDenyNone);
      if (not ZEXSLXReadContentTypes(stream,  FileArray, FilesCount)) then
      begin
        result := 3;
        exit;
      end;

      FreeAndNil(stream);

      ZE_XSLXReplaceDelimiter(FileArray, FilesCount);
      SheetRelationNumber := -1;

      b := false;
      for i := 0 to FilesCount - 1 do
      if (FileArray[i].ftype = 3) then
      begin
        b := true;
        break;
      end;

      if (not b) then
      begin
        s := DirName + '_rels' + PathDelim + '.rels';
        if (FileExists(s)) then
        begin
          SetLength(FileArray, FilesCount + 1);
          s := '/_rels/.rels';
          FileArray[FilesCount].original := s;
          FileArray[FilesCount].name := s;
          FileArray[FilesCount].ftype := 3;
          inc(FilesCount);
        end;

        s := DirName + 'xl' + PathDelim + '_rels' + PathDelim + 'workbook.xml.rels';
        if (FileExists(s)) then
        begin
          SetLength(FileArray, FilesCount + 1);
          s := '/xl/_rels/workbook.xml.rels';
          FileArray[FilesCount].original := s;
          FileArray[FilesCount].name := s;
          FileArray[FilesCount].ftype := 3;
          inc(FilesCount);
        end;

        ZE_XSLXReplaceDelimiter(FileArray, FilesCount);
      end;

      for i := 0 to FilesCount - 1 do
      if (FileArray[i].ftype = 3) then
      begin
        SetLength(RelationsArray, RelationsCount + 1);
        SetLength(RelationsCounts, RelationsCount + 1);

        stream := TFileStream.Create(DirName + FileArray[i].name, fmOpenRead or fmShareDenyNone);
        if (not ZE_XSLXReadRelationships(stream, RelationsArray[RelationsCount], RelationsCounts[RelationsCount], b, true)) then
        begin
          result := 4;
          exit;
        end;
        if (b) then
        begin
          SheetRelationNumber := RelationsCount;
          for j := 0 to RelationsCounts[RelationsCount] - 1 do
          if (RelationsArray[RelationsCount][j].ftype = 0) then
            for k := 0 to FilesCount - 1 do
            if (RelationsArray[RelationsCount][j].fileid < 0) then
              if ((pos(RelationsArray[RelationsCount][j].target, FileArray[k].original)) > 0) then
              begin
                RelationsArray[RelationsCount][j].fileid := k;
                break;
              end;
        end; //if
        FreeAndNil(stream);
        inc(RelationsCount);
      end;

      //sharedStrings.xml
      for i:= 0 to FilesCount - 1 do
      if (FileArray[i].ftype = 4) then
      begin
        FreeAndNil(stream);
        stream := TFileStream.Create(DirName + FileArray[i].name, fmOpenRead or fmShareDenyNone);
        if (not ZEXSLXReadSharedStrings(stream, StrArray, StrCount)) then
        begin
          result := 3;
          exit;
        end;
        break;
      end;

      // (styles.xml)
      for i := 0 to FilesCount - 1 do
      if (FileArray[i].ftype = 1) then
      begin
        FreeAndNil(stream);
        stream := TFileStream.Create(DirName + FileArray[i].name, fmOpenRead or fmShareDenyNone);
        if (not ZEXSLXReadStyles(XMLSS, stream)) then
        begin
          result := 5;
          exit;
        end else
          b := true;
      end;

      // 
      if (SheetRelationNumber > 0) then
      begin
        for i := 0 to FilesCount - 1 do
        if (FileArray[i].ftype = 2) then
        begin
          stream := TFileStream.Create(DirName + FileArray[i].name, fmOpenRead or fmShareDenyNone);
          if (not ZEXSLXReadWorkBook(XMLSS, stream, RelationsArray[SheetRelationNumber], RelationsCounts[SheetRelationNumber])) then
          begin
            result := 3;
            exit;
          end;
          break;
        end; //if

        for i := 1 to RelationsCounts[SheetRelationNumber] do
        for j := 0 to RelationsCounts[SheetRelationNumber] - 1 do
        if (RelationsArray[SheetRelationNumber][j].sheetid = i) then
        begin
          FreeAndNil(stream);
          stream := TFileStream.Create(DirName + FileArray[RelationsArray[SheetRelationNumber][j].fileid].name, fmOpenRead or fmShareDenyNone);
          if (not ZEXSLXReadSheet(XMLSS, stream, RelationsArray[SheetRelationNumber][j].name, StrArray, StrCount)) then
            result := result or 4;
          break;
        end; //if
      end else
      for i := 0 to FilesCount - 1 do
      if (FileArray[i].ftype = 0) then
      begin
        FreeAndNil(stream);
        stream := TFileStream.Create(DirName + FileArray[i].name, fmOpenRead or fmShareDenyNone);
        if (not ZEXSLXReadSheet(XMLSS, stream, '', StrArray, StrCount)) then
          result := result or 4;
      end;
    except
      result := 2;
    end;
  finally
    if (Assigned(stream)) then
      FreeAndNil(stream);

    SetLength(FileArray, 0);
    FileArray := nil;
    SetLength(StrArray, 0);
    StrArray := nil;
    for i := 0 to RelationsCount - 1 do
    begin
      Setlength(RelationsArray[i], 0);
      RelationsArray[i] := nil;
    end;
    SetLength(RelationsCounts, 0);
    RelationsCounts := nil;
    SetLength(RelationsArray, 0);
    RelationsArray := nil;
  end;
end; //ReadXSLXPath

{$IFDEF FPC}
// xlsx
//INPUT
//  var XMLSS: TZEXMLSS - 
//  FileName: string    -  
//RETURN
//      integer -   (0 -  OK)
function ReadXSLX(var XMLSS: TZEXMLSS; FileName: string): integer;
var
  i, j, k: integer;
  b: boolean;
  u_zip: TUnZipper;
  lst: TStringList;
  ZH: TXSLXZipHelper;

  procedure _getFile(num: integer);
  begin
    lst.Clear();
    lst.Add(ZH.FileArray[num].name); // 
    u_zip.UnZipFiles(lst);
  end;

begin
  result := 0;

  if (not FileExists(FileName)) then
  begin
    result := -1;
    exit;
  end;

  XMLSS.Styles.Clear();
  XMLSS.Sheets.Count := 0;

  u_zip := nil;
  lst := nil;
  ZH := nil;

  try
    try
      lst := TStringList.Create();
      lst.Clear();
      lst.Add('[Content_Types].xml'); // 
      ZH := TXSLXZipHelper.Create();
      ZH.XMLSS := XMLSS;
      u_zip := TUnZipper.Create();
      u_zip.FileName := FileName;
      u_zip.OnCreateStream := @ZH.DoCreateOutZipStream;
      u_zip.OnDoneStream := @ZH.DoDoneOutZipStream;
      ZH.FileType := -2;
      u_zip.UnZipFiles(lst);
      if (ZH.RetCode <> 0) then
      begin
        result := ZH.RetCode;
        exit;
      end;

      ZH.FileType := 3;
      for i := 0 to ZH.FilesCount - 1 do
      if (ZH.FileArray[i].ftype = 3) then
      begin
        ZH.FileNumber := i;
        _getFile(i);
        if (ZH.RetCode <> 0) then
        begin
          result := ZH.RetCode;
          exit;
        end;
      end;

      //sharedStrings.xml
      ZH.FileType := 4;
      for i:= 0 to ZH.FilesCount - 1 do
      if (ZH.FileArray[i].ftype = 4) then
      begin
        _getFile(i);
        if (ZH.RetCode <> 0) then
        begin
          result := ZH.RetCode;
          exit;
        end;
        break;
      end;

      // (styles.xml)
      ZH.FileType := 1;
      for i := 0 to ZH.FilesCount - 1 do
      if (ZH.FileArray[i].ftype = 1) then
      begin
        _getFile(i);
        if (ZH.RetCode <> 0) then
        begin
          result := ZH.RetCode;
          exit;
        end;
      end;

      // 
      if (ZH.SheetRelationNumber > 0) then
      begin
        ZH.FileType := 2;
        for i := 0 to ZH.FilesCount - 1 do
        if (ZH.FileArray[i].ftype = 2) then
        begin
          _getFile(i);
          if (ZH.RetCode <> 0) then
          begin
            result := ZH.RetCode;
            exit;
          end;
          break;
        end; //if

        ZH.FileType := 0;
        for i := 1 to ZH.RelationsCounts[ZH.SheetRelationNumber] do
        if (ZH.RelationsArray[ZH.SheetRelationNumber][i].sheetid = i) then
        begin
          ZH.ListName := ZH.RelationsArray[ZH.SheetRelationNumber][i].name;
          _getFile(ZH.RelationsArray[ZH.SheetRelationNumber][i].fileid);
          if (ZH.RetCode <> 0) then
            result := result or ZH.RetCode;
        end; //if
      end else
      for i := 0 to ZH.FilesCount - 1 do
      if (ZH.FileArray[i].ftype = 0) then
      begin
        _getFile(i);
        if (ZH.RetCode <> 0) then
          result := result or ZH.RetCode;
      end;
    except
      result := 2;
    end;
  finally
    if (Assigned(u_zip)) then
      FreeAndNil(u_zip);
    if (Assigned(lst)) then
       FreeAndNil(lst);
    if (Assigned(ZH)) then
       FreeAndNil(ZH);
  end;
end; //ReadXSLX
{$ENDIF}

/////////////////////////////////////////////////////////////////////////////
/////////////                                             /////////////
/////////////////////////////////////////////////////////////////////////////

// [Content_Types].xml 
//INPUT
//  var XMLSS: TZEXMLSS                 - 
//    Stream: TStream                   -   
//    TextConverter: TAnsiToCPConverter -      
//    PageCount: integer                - - 
//    CommentCount: integer             - -   
//  const PagesComments: TZESaveIntArray-     (  )
//    CodePageName: string              -   
//    BOM: ansistring                   - BOM
//RETURN
//      integer
function ZEXLSXCreateContentTypes(var XMLSS: TZEXMLSS; Stream: TStream; PageCount: integer; CommentCount: integer; const PagesComments: TZESaveIntArray;
                                  TextConverter: TAnsiToCPConverter; CodePageName: string; BOM: ansistring): integer;
var
  _xml: TZsspXMLWriterH;    //
  s: string;

  procedure _WriteOverride(const PartName: string; ct: integer);
  begin
    _xml.Attributes.Clear();
    _xml.Attributes.Add('PartName', PartName);
    case ct of
      0: s := 'application/vnd.openxmlformats-officedocument.spreadsheetml.worksheet+xml';
      1: s := 'application/vnd.openxmlformats-officedocument.spreadsheetml.styles+xml';
      2: s := 'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet.main+xml';
      3: s := 'application/vnd.openxmlformats-package.relationships+xml';
      4: s := 'application/vnd.openxmlformats-officedocument.spreadsheetml.sharedStrings+xml';
      5: s := 'application/vnd.openxmlformats-package.core-properties+xml';
      6: s := 'application/vnd.openxmlformats-officedocument.spreadsheetml.comments+xml';
      7: s := 'application/vnd.openxmlformats-officedocument.vmlDrawing';
    end;
    _xml.Attributes.Add('ContentType', s, false);
    _xml.WriteEmptyTag('Override', true);
  end; //_WriteOverride

  procedure _WriteTypes();
  var
    i: integer;

  begin
    //
    for i := 0 to PageCount - 1 do
      _WriteOverride('/xl/worksheets/sheet' + IntToStr(i + 1) + '.xml', 0);
    //  
    for i := 0 to CommentCount - 1 do
    begin
      _WriteOverride('/xl/worksheets/_rels/sheet' + IntToStr(PagesComments[i] + 1) + '.xml.rels', 3);
      _WriteOverride('/xl/comments' + IntToStr(PagesComments[i] + 1) + '.xml', 6);
    end;
  end; //_WriteTypes

begin
  result := 0;
  _xml := nil;
  try
    _xml := TZsspXMLWriterH.Create();
    _xml.TabLength := 1;
    _xml.TextConverter := TextConverter;
    _xml.TabSymbol := ' ';
    if (not _xml.BeginSaveToStream(Stream)) then
    begin
      result := 2;
      exit;
    end;

    ZEWriteHeaderCommon(_xml, CodePageName, BOM);
    _xml.Attributes.Clear();
    _xml.Attributes.Add('xmlns', 'http://schemas.openxmlformats.org/package/2006/content-types');
    _xml.WriteTagNode('Types', true, true, true);
    _WriteTypes();
    _xml.WriteEndTagNode(); //Types
  finally
    if (Assigned(_xml)) then
      FreeAndNil(_xml);
  end;
end; //ZEXLSXCreateContentTypes

//   (sheet*.xml)
//INPUT
//  var XMLSS: TZEXMLSS                 - 
//    Stream: TStream                   -   
//    SheetNum: integer                 -    
//    TextConverter: TAnsiToCPConverter -      
//    CodePageName: string              -   
//    BOM: ansistring                   - BOM
//  var isHaveComments: boolean         -  true,    (  comments*.xml)
//RETURN
//      integer
function ZEXLSXCreateSheet(var XMLSS: TZEXMLSS; Stream: TStream; SheetNum: integer; TextConverter: TAnsiToCPConverter; CodePageName: string; BOM: ansistring; var isHaveComments: boolean): integer;
var
  _xml: TZsspXMLWriterH;    //
  _sheet: TZSheet;

  procedure WriteXLSXSheetHeader();
  begin
    _xml.Attributes.Clear();
    _xml.Attributes.Add('filterMode', 'false');
    _xml.WriteTagNode('sheetPr', true, true, false);

    _xml.Attributes.Clear();
    _xml.Attributes.Add('fitToPage', 'false');
    _xml.WriteEmptyTag('pageSetUpPr', true, false);

    _xml.WriteEndTagNode(); //sheetPr

    _xml.Attributes.Clear();
    _xml.Attributes.Add('ref', 'A1:' + ZEGetA1byCol(_sheet.ColCount - 1) + IntToStr(_sheet.RowCount));
    _xml.WriteEmptyTag('dimension', true, false);

    _xml.Attributes.Clear();
    _xml.WriteTagNode('sheetViews', true, true, true);
    _xml.WriteEndTagNode(); //sheetViews
  end; //WriteXLSXSheetHeader

  procedure WriteXLSXSheetCols();
  var
    i: integer;
    s: string;

  begin
    _xml.Attributes.Clear();
    _xml.WriteTagNode('cols', true, true, true);
    for i := 0 to _sheet.ColCount - 1 do
    begin
      _xml.Attributes.Clear();
      _xml.Attributes.Add('collapsed', 'false', false);
      _xml.Attributes.Add('hidden', XLSXBoolToStr(_sheet.Columns[i].Hidden), false);
      _xml.Attributes.Add('max', '1', false); //??
      _xml.Attributes.Add('min', '1', false); //??
      s := '0';
      if ((_sheet.Columns[i].StyleID >= -1) and (_sheet.Columns[i].StyleID < XMLSS.Styles.Count)) then
        s := IntToStr(_sheet.Columns[i].StyleID + 1);
      _xml.Attributes.Add('style', s, false);
      _xml.Attributes.Add('width', ZEFloatSeparator(FormatFloat('0.##########', _sheet.Columns[i].WidthMM * 5.14509803921569 / 10)), false);
      _xml.WriteEmptyTag('col', true, false);
    end;
    _xml.WriteEndTagNode(); //cols
  end; //WriteXLSXSheetCols

  procedure WriteXLSXSheetData();
  var
    i, j, n: integer;
    b: boolean;
    s: string;
    _r: TRect;
    
  begin
    _xml.Attributes.Clear();
    _xml.WriteTagNode('sheetData', true, true, true);
    n := _sheet.ColCount - 1;
    for i := 0 to _sheet.RowCount - 1 do
    begin
      _xml.Attributes.Clear();
      _xml.Attributes.Add('collapsed', 'false', false); //?
      _xml.Attributes.Add('customFormat', 'false', false); //?
      _xml.Attributes.Add('customHeight', 'false', false); //?
      _xml.Attributes.Add('hidden', XLSXBoolToStr(_sheet.Rows[i].Hidden), false);
      _xml.Attributes.Add('ht', ZEFloatSeparator(FormatFloat('0.##', _sheet.Rows[i].HeightMM * 2.835)), false);
      _xml.Attributes.Add('outlineLevel', '0', false);
      _xml.Attributes.Add('r', IntToStr(i + 1), false);
      _xml.WriteTagNode('row', true, true, false);
      for j := 0 to n do
      begin
        _xml.Attributes.Clear();
        if (not isHaveComments) then
          if (length(_sheet.Cell[j, i].Comment) > 0) then
            isHaveComments := true;
        b := (length(_sheet.Cell[j, i].Data) > 0) or
             (length(_sheet.Cell[j, i].Formula) > 0);
        _xml.Attributes.Add('r', ZEGetA1byCol(j) + IntToStr(i + 1));
        if (_sheet.Cell[j, i].CellStyle >= -1) and (_sheet.Cell[j, i].CellStyle < XMLSS.Styles.Count) then
          s := IntToStr(_sheet.Cell[j, i].CellStyle + 1)
        else
          s := '0';
        _xml.Attributes.Add('s', s, false);

        case _sheet.Cell[j, i].CellType of
          ZENumber: s := 'n';
          ZEDateTime: s := 'str'; //??
          ZEBoolean: s := 'b';
          ZEansistring: s := 'str';
          ZEError: s := 'e';
        end;
        
        _xml.Attributes.Add('t', s, false);
        if (b) then
        begin
          _xml.WriteTagNode('c', true, true, false);
          if (length(_sheet.Cell[j, i].Formula) > 0) then
          begin
            _xml.Attributes.Clear();
            _xml.Attributes.Add('aca', 'false');
            _xml.WriteTag('f', _sheet.Cell[j, i].Formula, true, false, true);
          end;
          if (length(_sheet.Cell[j, i].Data) > 0) then
          begin
            _xml.Attributes.Clear();
            _xml.WriteTag('v', _sheet.Cell[j, i].Data, true, false, true);
          end;
          _xml.WriteEndTagNode();
        end else
          _xml.WriteEmptyTag('c', true);
      end;
      _xml.WriteEndTagNode(); //row
    end; //for i

    _xml.WriteEndTagNode(); //sheetData

    //Merge Cells
    if (_sheet.MergeCells.Count > 0) then
    begin
      _xml.Attributes.Clear();
      _xml.Attributes.Add('count', IntToStr(_sheet.MergeCells.Count));
      _xml.WriteTagNode('mergeCells', true, true, false);
      for i := 0 to _sheet.MergeCells.Count - 1 do
      begin
        _xml.Attributes.Clear();
        _r := _sheet.MergeCells.Items[i];
        s := ZEGetA1byCol(_r.Left) + IntToStr(_r.Top + 1) + ':' + ZEGetA1byCol(_r.Right) + IntToStr(_r.Bottom + 1);
        _xml.Attributes.Add('ref', s);
        _xml.WriteEmptyTag('mergeCell', true, false);
      end;
      _xml.WriteEndTagNode(); //mergeCells
    end; //if
  end; //WriteXLSXSheetData

  procedure WriteXLSXSheetFooter();
  var
    s: string;

  begin
    _xml.Attributes.Clear();
    _xml.Attributes.Add('headings', 'false', false);
    _xml.Attributes.Add('gridLines', 'false', false);
    _xml.Attributes.Add('gridLinesSet', 'true', false);
    _xml.Attributes.Add('horizontalCentered', XLSXBoolToStr(_sheet.SheetOptions.CenterHorizontal), false);
    _xml.Attributes.Add('verticalCentered', XLSXBoolToStr(_sheet.SheetOptions.CenterVertical), false);
    _xml.WriteEmptyTag('printOptions', true, false);

    _xml.Attributes.Clear();
    s := '0.##########';
    _xml.Attributes.Add('left',   ZEFloatSeparator(FormatFloat(s, _sheet.SheetOptions.MarginLeft / ZE_MMinInch)),   false);
    _xml.Attributes.Add('right',  ZEFloatSeparator(FormatFloat(s, _sheet.SheetOptions.MarginRight / ZE_MMinInch)),  false);
    _xml.Attributes.Add('top',    ZEFloatSeparator(FormatFloat(s, _sheet.SheetOptions.MarginTop / ZE_MMinInch)),    false);
    _xml.Attributes.Add('bottom', ZEFloatSeparator(FormatFloat(s, _sheet.SheetOptions.MarginBottom / ZE_MMinInch)), false);
    _xml.Attributes.Add('header', ZEFloatSeparator(FormatFloat(s, _sheet.SheetOptions.HeaderMargin / ZE_MMinInch)), false);
    _xml.Attributes.Add('footer', ZEFloatSeparator(FormatFloat(s, _sheet.SheetOptions.FooterMargin / ZE_MMinInch)), false);
    _xml.WriteEmptyTag('pageMargins', true, false);

    _xml.Attributes.Clear();
    _xml.Attributes.Add('blackAndWhite', 'false', false);
    _xml.Attributes.Add('cellComments', 'none', false);
    _xml.Attributes.Add('copies', '1', false);
    _xml.Attributes.Add('draft', 'false', false);
    _xml.Attributes.Add('firstPageNumber', '1', false);
    _xml.Attributes.Add('fitToHeight', '1', false);
    _xml.Attributes.Add('fitToWidth', '1', false);
    _xml.Attributes.Add('horizontalDpi', '300', false);
    if (_sheet.SheetOptions.PortraitOrientation) then
      s := 'portrait'
    else
      s := 'album';
    _xml.Attributes.Add('orientation', s, false);
    _xml.Attributes.Add('pageOrder', 'downThenOver', false);
    _xml.Attributes.Add('paperSize', intToStr(_sheet.SheetOptions.PaperSize), false);
    _xml.Attributes.Add('scale', '100', false);
    _xml.Attributes.Add('useFirstPageNumber', 'true', false);
    _xml.Attributes.Add('usePrinterDefaults', 'false', false);
    _xml.Attributes.Add('verticalDpi', '300', false);
    _xml.WriteEmptyTag('pageSetup', true, false);
    //  <headerFooter differentFirst="false" differentOddEven="false">
    //    <oddHeader> ... </oddHeader>
    //    <oddFooter> ... </oddFooter>
    //  </headerFooter>
    //  <legacyDrawing r:id="..."/>
  end; //WriteXLSXSheetFooter

begin
  isHaveComments := false;
  result := 0;
  _xml := nil;
  try
    _xml := TZsspXMLWriterH.Create();
    _xml.TabLength := 1;
    _xml.TextConverter := TextConverter;
    _xml.TabSymbol := ' ';
    if (not _xml.BeginSaveToStream(Stream)) then
    begin
      result := 2;
      exit;
    end;

    ZEWriteHeaderCommon(_xml, CodePageName, BOM);

    _xml.Attributes.Clear();
    _xml.Attributes.Add('xmlns', 'http://schemas.openxmlformats.org/spreadsheetml/2006/main');
    _xml.Attributes.Add('xmlns:r', 'http://schemas.openxmlformats.org/officeDocument/2006/relationships');
    _xml.WriteTagNode('worksheet', true, true, false);

    _sheet := XMLSS.Sheets[SheetNum];
    WriteXLSXSheetHeader();
    WriteXLSXSheetCols();
    WriteXLSXSheetData();
    WriteXLSXSheetFooter();

    _xml.WriteEndTagNode(); //worksheet
  finally
    if (Assigned(_xml)) then
      FreeAndNil(_xml);
  end;
end; //ZEXLSXCreateSheet

// workbook.xml
//INPUT
//  var XMLSS: TZEXMLSS                 - 
//    Stream: TStream                   -   
//  const _pages: TZESaveIntArray       -  
//  const _names: TZESaveStrArray       -   
//    PageCount: integer                - - 
//    TextConverter: TAnsiToCPConverter -      
//    CodePageName: string              -   
//    BOM: ansistring                   - BOM
//RETURN
//      integer
function ZEXLSXCreateWorkBook(var XMLSS: TZEXMLSS; Stream: TStream; const _pages: TZESaveIntArray;
                              const _names: TZESaveStrArray; PageCount: integer; TextConverter: TAnsiToCPConverter; CodePageName: string; BOM: ansistring): integer;
var
  _xml: TZsspXMLWriterH;    //

  //<sheets> ... </sheets>
  procedure WriteSheets();
  var
    i: integer;

  begin
    _xml.Attributes.clear();
    _xml.WriteTagNode('sheets', true, true, true);

    for i := 0 to PageCount - 1 do
    begin
      _xml.Attributes.Clear();
      _xml.Attributes.Add('name', _names[i], false);
      _xml.Attributes.Add('sheetId', IntToStr(i + 1), false);
      _xml.Attributes.Add('state', 'visible', false);
      _xml.Attributes.Add('r:id', 'rId' + IntToStr(i + 2), false);
      _xml.WriteEmptyTag('sheet', true);
    end; //for i

    _xml.WriteEndTagNode(); //sheets
  end; //WriteSheets

begin
  result := 0;
  _xml := nil;
  try
    _xml := TZsspXMLWriterH.Create();
    _xml.TabLength := 1;
    _xml.TextConverter := TextConverter;
    _xml.TabSymbol := ' ';
    if (not _xml.BeginSaveToStream(Stream)) then
    begin
      result := 2;
      exit;
    end;

    ZEWriteHeaderCommon(_xml, CodePageName, BOM);

    _xml.Attributes.Clear();
    _xml.Attributes.Add('xmlns', 'http://schemas.openxmlformats.org/spreadsheetml/2006/main');
    _xml.Attributes.Add('xmlns:r', 'http://schemas.openxmlformats.org/officeDocument/2006/relationships', false);
    _xml.WriteTagNode('workbook', true, true, true);

    _xml.Attributes.Clear();
    _xml.Attributes.Add('appName', 'ZEXMLSSlib');
    _xml.WriteEmptyTag('fileVersion', true);

    _xml.Attributes.Clear();
    _xml.Attributes.Add('backupFile', 'false');
    _xml.Attributes.Add('showObjects', 'all', false);
    _xml.Attributes.Add('date1904', 'false', false);
    _xml.WriteEmptyTag('workbookPr', true);

    _xml.Attributes.Clear();
    _xml.WriteEmptyTag('workbookProtection', true);

    _xml.WriteTagNode('bookViews', true, true, true);

    _xml.Attributes.Add('activeTab', '0');
    _xml.Attributes.Add('firstSheet', '0', false);
    _xml.Attributes.Add('showHorizontalScroll', 'true', false);
    _xml.Attributes.Add('showSheetTabs', 'true', false);
    _xml.Attributes.Add('showVerticalScroll', 'true', false);
    _xml.Attributes.Add('tabRatio', '200', false);
    _xml.Attributes.Add('windowHeight', '8192', false);
    _xml.Attributes.Add('windowWidth', '16384', false);
    _xml.Attributes.Add('xWindow', '0', false);
    _xml.Attributes.Add('yWindow', '0', false);
    _xml.WriteEmptyTag('workbookView', true);
    _xml.WriteEndTagNode(); // bookViews

    WriteSheets();

    _xml.Attributes.Clear();
    _xml.Attributes.Add('iterateCount', '100');
    _xml.Attributes.Add('refMode', 'A1', false); //{tut}
    _xml.Attributes.Add('iterate', 'false', false);
    _xml.Attributes.Add('iterateDelta', '0.001', false);
    _xml.WriteEmptyTag('calcPr', true);

    _xml.WriteEndTagNode(); //workbook
  finally
    if (Assigned(_xml)) then
      FreeAndNil(_xml);
  end;
end; //ZEXLSXCreateWorkBook

// styles.xml 
//INPUT
//  var XMLSS: TZEXMLSS                 - 
//    Stream: TStream                   -   
//    TextConverter: TAnsiToCPConverter -      
//    CodePageName: string              -   
//    BOM: ansistring                   - BOM
//RETURN
//      integer
function ZEXLSXCreateStyles(var XMLSS: TZEXMLSS; Stream: TStream; TextConverter: TAnsiToCPConverter; CodePageName: string; BOM: ansistring): integer;
var
  _xml: TZsspXMLWriterH;        //
  _FontIndex: TZESaveIntArray;  // 
  _FillIndex: TZESaveIntArray;  //
  _BorderIndex: TZESaveIntArray;//
  _StylesCount: integer;        

  //   
  function _isFontsEqual(const fnt1, fnt2: TFont): boolean;
  begin
    result := true;
    if (fnt1.Color <> fnt2.Color) then
    begin
      result := false;
      exit;
    end;

    if (fnt1.Height <> fnt2.Height) then
    begin
      result := false;
      exit;
    end;

    if (fnt1.Name <> fnt2.Name) then
    begin
      result := false;
      exit;
    end;

    if (fnt1.Pitch <> fnt2.Pitch) then
    begin
      result := false;
      exit;
    end;

    if (fnt1.Size <> fnt2.Size) then
    begin
      result := false;
      exit;
    end;

    if (fnt1.Style <> fnt2.Style) then
    begin
      result := false;
      exit;
    end;
  end; //_isFontsEqual

  //   
  //INPUT
  //  var arr: TZESaveIntArray  - 
  //      cnt: integer          -      (  0)
  procedure _UpdateArrayIndex(var arr: TZESaveIntArray; cnt: integer);
  var
    i, j, n: integer;
    num: integer;

  begin
    num := 0;
    for i := -1 to cnt do
    if (arr[i + 1] = -2) then
    begin
      n := i + 1;
      for j := n to cnt do
      if (arr[j + 1] = i) then
        arr[j + 1] := num;
      arr[i + 1] := num;
      inc(num);
    end; //if
  end; //_UpdateArrayIndex

  //<fonts>...</fonts>
  procedure WriteXLSXFonts();
  var
    i, j, n: integer;
    _fontCount: integer;
    fnt: TFont;

  begin
    _fontCount := 0;
    SetLength(_FontIndex, _StylesCount + 1);
    for i := 0 to _StylesCount do
      _FontIndex[i] := -2;

    for i := -1 to _StylesCount - 1 do
    if (_FontIndex[i + 1] = -2) then
    begin
      inc (_fontCount);
      n := i + 1;
      for j := n to _StylesCount - 1 do
      if (_FontIndex[j + 1] = -2) then
        if (_isFontsEqual(XMLSS.Styles[n].Font, XMLSS.Styles[j].Font)) then
          _FontIndex[j + 1] := i;
    end; //if

    _xml.Attributes.Clear();
    _xml.Attributes.Add('count', IntToStr(_fontCount));
    _xml.WriteTagNode('fonts', true, true, true);

    for i := 0 to _StylesCount do
    if (_FontIndex[i] = - 2) then
    begin
      fnt := XMLSS.Styles[i - 1].Font;
      _xml.Attributes.Clear();
      _xml.WriteTagNode('font', true, true, true);

      _xml.Attributes.Clear();
      _xml.Attributes.Add('val', fnt.Name);
      _xml.WriteEmptyTag('name', true);

      _xml.Attributes.Clear();
      _xml.Attributes.Add('val', IntToStr(fnt.Charset));
      _xml.WriteEmptyTag('charset', true);

      _xml.Attributes.Clear();
      _xml.Attributes.Add('val', IntToStr(fnt.Size));
      _xml.WriteEmptyTag('sz', true);

      if (fnt.Color <> clWindowText) then
      begin
        _xml.Attributes.Clear();
        _xml.Attributes.Add('rgb', '00' + ColorToHTMLHex(fnt.Color));
        _xml.WriteEmptyTag('color', true);
      end;

      if (fsBold in fnt.Style) then
      begin
        _xml.Attributes.Clear();
        _xml.Attributes.Add('val', 'true');
        _xml.WriteEmptyTag('b', true);
      end;

      if (fsItalic in fnt.Style) then
      begin
        _xml.Attributes.Clear();
        _xml.Attributes.Add('val', 'true');
        _xml.WriteEmptyTag('i', true);
      end;

      if (fsStrikeOut in fnt.Style) then
      begin
        _xml.Attributes.Clear();
        _xml.Attributes.Add('val', 'true');
        _xml.WriteEmptyTag('strike', true);
      end;

      if (fsUnderline in fnt.Style) then
      begin
        _xml.Attributes.Clear();
        _xml.Attributes.Add('val', 'single');
        _xml.WriteEmptyTag('u', true);
      end;

      _xml.WriteEndTagNode(); //font
    end; //if

    _UpdateArrayIndex(_FontIndex, _StylesCount - 1);

    _xml.WriteEndTagNode(); //fonts
  end; //WriteXLSXFonts

  //   
  function _isFillsEqual(style1, style2: TZStyle): boolean;
  begin
    result := (style1.BGColor = style2.BGColor) and
              (style1.PatternColor = style2.PatternColor) and
              (style1.CellPattern = style2.CellPattern);
  end; //_isFillsEqual

  //<fills> ... </fills>
  procedure WriteXLSXFills();
  var
    i, j: integer;
    _fillCount: integer;
    s: string;

  begin
    _fillCount := 0;
    SetLength(_FillIndex, _StylesCount + 1);
    for i := -1 to _StylesCount - 1 do
      _FillIndex[i + 1] := -2;
    for i := -1 to _StylesCount - 1 do
    if (_FillIndex[i + 1] = - 2) then
    begin
      inc(_fillCount);
      for j := i + 1 to _StylesCount - 1 do
      if (_FillIndex[j + 1] = -2) then
        if (_isFillsEqual(XMLSS.Styles[i], XMLSS.Styles[j])) then
          _FillIndex[j + 1] := i;
    end; //if

    _xml.Attributes.Clear();
    _xml.Attributes.Add('count', IntToStr(_fillCount));
    _xml.WriteTagNode('fills', true, true, true);

    for i := -1 to _StylesCount - 1 do
    if (_FillIndex[i + 1] = -2) then
    begin
      _xml.Attributes.Clear();
      _xml.WriteTagNode('fill', true, true, true);

      case XMLSS.Styles[i].CellPattern of
        ZPSolid: s := 'solid';
        ZPNone: s := 'none';
        ZPGray125: s := 'gray125';
        ZPGray0625: s := 'gray0625';
        else
          s := 'solid';
      end; //case
      _xml.Attributes.Clear();
      _xml.Attributes.Add('patternType', s);
      _xml.WriteEmptyTag('patternFill', true);

      if (XMLSS.Styles[i].PatternColor <> clWindow) then
      begin
        _xml.Attributes.Clear();
        _xml.Attributes.Add('rgb', '00' + ColorToHTMLHex(XMLSS.Styles[i].PatternColor));
        _xml.WriteEmptyTag('fgColor', true);
      end;

      if (XMLSS.Styles[i].BGColor <> clWindow) then
      begin
        _xml.Attributes.Clear();
        _xml.Attributes.Add('rgb', '00' + ColorToHTMLHex(XMLSS.Styles[i].BGColor));
        _xml.WriteEmptyTag('bgColor', true);
      end;

      _xml.WriteEndTagNode(); //fill
    end; //if

    _UpdateArrayIndex(_FillIndex, _StylesCount - 1);

    _xml.WriteEndTagNode(); //fills
  end; //WriteXLSXFills();

  // 
  procedure _WriteBorderItem(StyleNum: integer; BorderNum: integer);
  var
    s, s1: string;
    _border: TZBorderStyle;
    n: integer;

  begin
    _xml.Attributes.Clear();
    case BorderNum of
      0: s := 'left';
      1: s := 'top';
      2: s := 'right';
      3: s := 'bottom';
      else
        s := 'diagonal';
    end;
    _border := XMLSS.Styles[StyleNum].Border[BorderNum];
    s1 := '';
    case _border.LineStyle of
      ZEContinuous:
        begin
          //thin ??
          if (_border.Weight = 1) then
            s1 := 'hair'
          else
          if (_border.Weight = 2) then
            s1 := 'medium'
          else
            s1 := 'thick';
        end;
      ZEDash:
        begin
          if (_border.Weight = 1) then
            s1 := 'dashed'
          else
          if (_border.Weight >= 2) then
            s1 := 'mediumDashed';
        end;
      ZEDot:
        begin
          if (_border.Weight = 1) then
            s1 := 'dotted'
          else
          if (_border.Weight >= 2) then
            s1 := 'mediumDotted';
        end;
      ZEDashDot:
        begin
          if (_border.Weight = 1) then
            s1 := 'dashDot'
          else
          if (_border.Weight >= 2) then
            s1 := 'mediumDashDot';
        end;
      ZEDashDotDot:
        begin
          if (_border.Weight = 1) then
            s1 := 'dashDotDot'
          else
          if (_border.Weight >= 2) then
            s1 := 'mediumDashDotDot';
        end;
      ZESlantDashDot:
        begin
          s1 := 'slantDashDot';
        end;
      ZEDouble:
        begin
          s1 := 'double';
        end;
      ZENone:
        begin
        end;
    end; //case

    n := length(s1);

    if (n > 0) then
      _xml.Attributes.Add('style', s1);

    if ((_border.Color <> clBlack) and (n > 0)) then
    begin
      _xml.WriteTagNode(s, true, true, true);
      _xml.Attributes.Clear();
      _xml.Attributes.Add('rgb', '00' + ColorToHTMLHex(_border.Color));
      _xml.WriteEmptyTag('color', true);
      _xml.WriteEndTagNode();
    end else
      _xml.WriteEmptyTag(s, true);
  end; //_WriteBorderItem

  //<borders> ... </borders>
  procedure WriteXLSXBorders();
  var
    i, j: integer;
    _borderCount: integer;
    s: string;

  begin
    _borderCount := 0;
    SetLength(_BorderIndex, _StylesCount + 1);
    for i := -1 to _StylesCount - 1 do
      _BorderIndex[i + 1] := -2;
    for i := -1 to _StylesCount - 1 do
    if (_BorderIndex[i + 1] = - 2) then
    begin
      inc(_borderCount);
      for j := i + 1 to _StylesCount - 1 do
      if (_BorderIndex[j + 1] = -2) then
        if (XMLSS.Styles[i].IsEqual(XMLSS.Styles[j])) then
          _BorderIndex[j + 1] := i;
    end; //if

    _xml.Attributes.Clear();
    _xml.Attributes.Add('count', IntToStr(_borderCount));
    _xml.WriteTagNode('borders', true, true, true);

    for i := -1 to _StylesCount - 1 do
    if (_BorderIndex[i + 1] = -2) then
    begin
      _xml.Attributes.Clear();
      s := 'false';
      if (XMLSS.Styles[i].Border[4].Weight > 0) and (XMLSS.Styles[i].Border[4].LineStyle <> ZENone) then
        s := 'true';
      _xml.Attributes.Add('diagonalDown', s);
      s := 'false';
      if (XMLSS.Styles[i].Border[5].Weight > 0) and (XMLSS.Styles[i].Border[5].LineStyle <> ZENone) then
        s := 'true';
      _xml.Attributes.Add('diagonalUp', s, false);
      _xml.WriteTagNode('border', true, true, true);

      for j := 0 to 3 do
        _WriteBorderItem(i, j);

      _xml.WriteEndTagNode(); //border
    end; //if

    _UpdateArrayIndex(_BorderIndex, _StylesCount - 1);

    _xml.WriteEndTagNode(); //borders
  end; //WriteXLSXBorders

  // <xf> ... </xf>
  //INPUT
  //      NumStyle: integer -  
  //      isxfId: boolean   -     "xfId"
  //      xfId: integer     -  "xfId"
  procedure _WriteXF(NumStyle: integer; isxfId: boolean; xfId: integer);
  var
    _addalignment: boolean;
    _style: TZStyle;
    s: string;

  begin
    _xml.Attributes.Clear();
    _style := XMLSS.Styles[NumStyle];
    _addalignment := _style.Alignment.WrapText or
                    (_style.Alignment.Rotate <> 0) or
                    (_style.Alignment.Indent <> 0) or
                    _style.Alignment.ShrinkToFit or
                    (_style.Alignment.Vertical <> ZVAutomatic) or
                    (_style.Alignment.Horizontal <> ZHAutomatic);

    _xml.Attributes.Add('applyAlignment', XLSXBoolToStr(_addalignment));
    _xml.Attributes.Add('applyBorder', 'true', false);
    _xml.Attributes.Add('applyFont', 'true', false);
    _xml.Attributes.Add('applyProtection', 'true', false);
    _xml.Attributes.Add('borderId', IntToStr(_BorderIndex[NumStyle + 1]), false);
    _xml.Attributes.Add('fillId', IntToStr(_FillIndex[NumStyle + 1]), false);
    _xml.Attributes.Add('fontId', IntToStr(_FontIndex[NumStyle + 1]), false);
    _xml.Attributes.Add('numFmtId', '164', false); //??

    if (isxfId) then
      _xml.Attributes.Add('xfId', IntToStr(xfId), false);

    _xml.WriteTagNode('xf', true, true, true);

    if (_addalignment) then
    begin
      _xml.Attributes.Clear();
      case (_style.Alignment.Horizontal) of
        ZHLeft: s := 'left';
        ZHRight: s := 'right';
        ZHCenter: s := 'center';
        ZHFill: s := 'fill';
        ZHJustify: s := 'justify';
        ZHDistributed: s := 'distributed';
        else
          s := 'center';
      end; //case
      _xml.Attributes.Add('horizontal', s);
      _xml.Attributes.Add('indent', IntToStr(_style.Alignment.Indent), false);
      _xml.Attributes.Add('shrinkToFit', XLSXBoolToStr(_style.Alignment.ShrinkToFit), false);
      _xml.Attributes.Add('textRotation', IntToStr(_style.Alignment.Rotate), false);
      case (_style.Alignment.Vertical) of
        ZVCenter: s := 'center';
        ZVTop: s := 'top';
        ZVBottom: s := 'bottom';
        ZVJustify: s := 'justify';
        ZVDistributed: s := 'distributed';
        else
          s := 'center';
      end; //case
      _xml.Attributes.Add('vertical', s, false);
      _xml.Attributes.Add('wrapText', XLSXBoolToStr(_style.Alignment.WrapText), false);
      _xml.WriteEmptyTag('alignment', true);
    end; //if (_addalignment)

    _xml.Attributes.Clear();
    _xml.Attributes.Add('hidden', XLSXBoolToStr(XMLSS.Styles[NumStyle].Protect));
    _xml.Attributes.Add('locked', XLSXBoolToStr(XMLSS.Styles[NumStyle].HideFormula));
    _xml.WriteEmptyTag('protection', true);

    _xml.WriteEndTagNode(); //xf
  end; //_WriteXF

  //<cellStyleXfs> ... </cellStyleXfs> / <cellXfs> ... </cellXfs>
  procedure WriteCellStyleXfs(const TagName: string; isxfId: boolean);
  var
    i: integer;

  begin
    _xml.Attributes.Clear();
    _xml.Attributes.Add('count', IntToStr(XMLSS.Styles.Count + 1));
    _xml.WriteTagNode(TagName, true, true, true);
    for i := -1 to XMLSS.Styles.Count - 1 do
    begin
      _WriteXF(i, isxfId, i + 1);
    end;
    _xml.WriteEndTagNode(); //cellStyleXfs
  end; //WriteCellStyleXfs

  //<cellStyles> ... </cellStyles>
  procedure WriteCellStyles();
  begin
  end; //WriteCellStyles 

begin
  result := 0;
  _xml := nil;
  try
    _xml := TZsspXMLWriterH.Create();
    _xml.TabLength := 1;
    _xml.TextConverter := TextConverter;
    _xml.TabSymbol := ' ';
    if (not _xml.BeginSaveToStream(Stream)) then
    begin
      result := 2;
      exit;
    end;

    ZEWriteHeaderCommon(_xml, CodePageName, BOM);
    _StylesCount := XMLSS.Styles.Count;

    _xml.Attributes.Clear();
    _xml.Attributes.Add('xmlns', 'http://schemas.openxmlformats.org/spreadsheetml/2006/main');
    _xml.WriteTagNode('styleSheet', true, true, true);

    WriteXLSXFonts();
    WriteXLSXFills();
    WriteXLSXBorders();
    WriteCellStyleXfs('cellStyleXfs', false);
    WriteCellStyleXfs('cellXfs', true);
    WriteCellStyles(); //??

    _xml.WriteEndTagNode(); //styleSheet
  finally
    if (Assigned(_xml)) then
      FreeAndNil(_xml);
    SetLength(_FontIndex, 0);
    _FontIndex := nil;
    SetLength(_FillIndex, 0);
    _FillIndex := nil;
    SetLength(_BorderIndex, 0);
    _BorderIndex := nil;
  end;
end; //ZEXLSXCreateStyles

//     Office Open XML (OOXML)
//INPUT
//  var XMLSS: TZEXMLSS                   - 
//      PathName: string                  -      (   )
//  const SheetsNumbers:array of integer  -      
//  const SheetsNames: array of string    -   
//                                          (      )
//      TextConverter: TAnsiToCPConverter - 
//      CodePageName: string              -  
//      BOM: ansistring                   - Byte Order Mark
//RETURN
//      integer
function SaveXmlssToXLSXPath(var XMLSS: TZEXMLSS; PathName: string; const SheetsNumbers:array of integer;
                         const SheetsNames: array of string; TextConverter: TAnsiToCPConverter; CodePageName: string; BOM: ansistring = ''): integer;
var
  _pages: TZESaveIntArray;      // 
  _names: TZESaveStrArray;      // 
  kol: integer;
  Stream: TStream;
  need_comments: boolean;

begin
  result := 0;
  Stream := nil;
  try
    if (not ZE_CheckDirExist(PathName)) then
    begin
      result := 3;
      exit;
    end;

    if (not ZECheckTablesTitle(XMLSS, SheetsNumbers, SheetsNames, _pages, _names, kol)) then
    begin
      result := 2;
      exit;
    end;


    Stream := TFileStream.Create(PathName + 'sheet1.xml', fmCreate);
    ZEXLSXCreateSheet(XMLSS, Stream, 0, TextConverter, CodePageName, BOM, need_comments);
    FreeAndNil(Stream);

    Stream := TFileStream.Create(PathName + 'styles1.xml', fmCreate);
    ZEXLSXCreateStyles(XMLSS, Stream, TextConverter, CodePageName, BOM);
    FreeAndNil(Stream);


  finally
    ZESClearArrays(_pages, _names);
    if (Assigned(Stream)) then
      FreeAndNil(Stream);
  end;
end; //SaveXmlssToXLSXPath                        

end.
